"""**Tools for creating patterns of geometric shapes**

The :py:mod:`mahautils.shapes` module is intended to facilitate building
complex patterns of geometric shapes.  This can be useful when preparing
input files for simulations, such as files that specify port shapes and
locations for pumps and motors.
"""

from .canvas import Canvas
from .geometry import (
    CartesianPoint2D,
    CartesianPoint3D,
    Circle,
    ClosedShape2D,
    Geometry,
    OpenShape2D,
    Point,
    Polygon,
    Shape2D,
)
from .layer import Layer



"""Runs MahaUtils CLI and GUI tools

This module allows MahaUtils tools to be run from the command line using:

.. code-block:: shell

    python3 -m mahautils [COMMAND]

Where ``[COMMAND]`` is one of the following:

- ``simviewer``: Simulation results viewer
"""

import sys

from .multics.sim_results_viewer.app import main as run_sim_results_viewer


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('MahaUtils: error: No command provided')
        sys.exit(1)

    command = sys.argv[1]
    if command.lower() == 'simviewer':
        run_sim_results_viewer(sys.argv[2:])
    else:
        print(f'MahaUtils: error: Invalid command "{command}"')



"""Utilities for managing data stored in arrays.
"""

from typing import Any

import numpy as np


def to_np_1D_array(value: Any, dtype: Any = None,
                   flatten: bool = True) -> np.ndarray:
    """Converts an input to a 1D NumPy array

    This function converts data of any type into a 1D NumPy array.

    Parameters
    ----------
    value : Any
        The data to convert to a 1D NumPy array
    dtype : Any, optional
        The required data type for the items in the NumPy array to be returned
        (default is ``None``, which imposes no data type requirement).  Any
        valid data type accepted by the ``numpy.array()`` function can be
        provided
    flatten : bool, optional
        Whether to use NumPy's `flatten <https://numpy.org/doc/stable/reference
        /generated/numpy.ndarray.flatten.html>`__ function to convert ``value``
        to a 1D array in the event ``value`` has two or more dimensions
        (default is ``True``)

    Returns
    -------
    np.ndarray
        The data in ``value``, converted to a 1D NumPy array
    """
    np_array = np.array(value, dtype=dtype)

    if flatten:
        np_array = np_array.flatten()

    if np_array.ndim == 0:
        np_array = np.expand_dims(np_array, axis=0)

    if np_array.ndim != 1:
        raise ValueError('Input must be at most 1D, but given '
                         f'data are {np_array.ndim}D')

    return np_array



"""Code for the components of the simulation results viewer that configure
data and settings.
"""

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash_bootstrap_components as dbc  # type: ignore

from mahautils.multics.sim_results_viewer.constants import (
    GUI_SHORT_NAME,
)
from .tab_data import data_files_tab
from .tab_plot_settings import plot_settings_tab


def simviewer_config_panel():
    """Creates the slide-in panel where users can configure SimViewer settings
    """
    default_width_percent = 45
    default_open = True

    contents = dbc.Tabs([
        dbc.Tab(
            data_files_tab(),
            label='Data Files',
            active_tab_class_name='fw-bold',
        ),
        dbc.Tab(
            plot_settings_tab(),
            label='Plot Settings',
            active_tab_class_name='fw-bold',
        ),
    ])

    return dbc.Offcanvas(
        contents,
        id='plot-config-panel',
        title=f'{GUI_SHORT_NAME} Configuration',
        placement='end',
        is_open=default_open,
        style={
            'width': f'{default_width_percent}vw',
        },
    )



"""User interface (UI) elements used throughout the SimViewer interface
"""

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash                              # type: ignore
import dash_bootstrap_components as dbc  # type: ignore

from .constants import COLOR_PICKER_STYLE, UI_DESCRIPTION_MARGIN_BELOW


## COLOR PICKER ##

def color_picker(identifier, current_color: str):
    """An interactive element that allows the user to choose a color and
    displays the currently selected color"""
    return dbc.Row([
        dbc.Col(
            dbc.Input(
                id=identifier,
                type='color',
                value=current_color,
                style=COLOR_PICKER_STYLE,
            ),
            width=3,
        ),
        dbc.Col([
            dash.html.P(
                'Selected color hex code: ',
                style={'fontWeight': 'bold', 'marginBottom': '0px'}),
            dash.html.P(
                current_color,
                style={'fontFamily': 'monospace', 'marginTop': '0px'}),
        ]),
    ])


## NUMBERED ITEM SELECTOR ##

_selector_button_style = {
    'borderRadius': '8px',
    'marginTop': '2px',
    'marginLeft': '6px',
    'marginRight': '0px',
    'marginBottom': '2px',
    'width': '45px',
}


def numbered_item_selector(pagination_id, add_button_id,
                           hide_show_button_id, delete_button_id,
                           name: str, name_plural: str,
                           num_items: int, active_page: int,
                           is_active_page_shown: bool,
                           disable_hide_show_button: bool = False):
    """A set of elements for creating and managing a numbered set of items

    This function creates a :py:class:`dbc.Pagination` object that can be
    used to select numbered items, as well as associated buttons for adding,
    deleting, and hiding/showing items.
    """
    axis_exists = (num_items != 0)

    return dash.html.Div([
        dbc.Row([
            dbc.Col(
                [
                    dash.html.Div(
                        dbc.Pagination(
                            id=pagination_id,
                            min_value=1, max_value=num_items,
                            fully_expanded=False,
                            first_last=False,
                            previous_next=True,
                            active_page=active_page,
                        ),
                        hidden=(not axis_exists),
                    ),
                    dash.html.Div(
                        dash.html.P(
                            f'No {name_plural} created yet.  Click the "+" '
                            f'button to right to create a new {name}.'),
                        hidden=axis_exists,
                    ),
                ],
                xs=4, sm=4, md=5, lg=6, xl=7, xxl=8,
            ),
            dbc.Col(
                dash.html.Div(
                    [
                        dbc.Button(
                            dash.html.I(className='fa-sharp fa-solid fa-plus'),
                            id=add_button_id,
                            size='md',
                            n_clicks=0,
                            style=_selector_button_style,
                        ),
                        dbc.Button(
                            dash.html.I(
                                className=(
                                    'fa fa-eye-slash' if is_active_page_shown
                                    else 'fa fa-eye'),
                            ),
                            id=hide_show_button_id,
                            size='md',
                            n_clicks=0,
                            style=_selector_button_style,
                            disabled=(
                                disable_hide_show_button or (not axis_exists)
                            ),
                        ),
                        dbc.Button(
                            dash.html.I(className='bi bi-trash'),
                            id=delete_button_id,
                            color='danger',
                            size='md',
                            n_clicks=0,
                            style=_selector_button_style,
                            disabled=(not axis_exists),
                        ),
                    ],
                    style={'textAlign': 'right'}
                ),
            ),
        ], style={'marginBottom': '0px'}),
        dbc.Tooltip(f'Select a {name} to edit', target=pagination_id,
                    trigger='hover'),
        dbc.Tooltip(f'Add a new {name}', target=add_button_id, trigger='hover'),
        dbc.Tooltip(f'{"Hide" if is_active_page_shown else "Show"} the '
                    f'currently selected {name}',
                    target=hide_show_button_id, trigger='hover'),
        dbc.Tooltip(f'Delete the currently selected {name}',
                    target=delete_button_id,
                    trigger='hover'),
        dash.html.Hr(style={'marginTop': '0px'}),
    ])


## ON/OFF SWITCH ##

def on_off_switch_row(identifier, state: bool, description: str):
    """A row with a single on/off switch and a description"""
    return dbc.Row([
        dbc.Col(
            dbc.Switch(
                id=identifier,
                value=state,
            ),
            width=2,
        ),
        dbc.Col(
            dash.html.P(description),
            width=10,
        ),
    ])


## TICK MARK OPTIONS ##

def tick_range_step(range_min_id, range_max_id, tick_spacing_id,
                    range_min_val, range_max_val, tick_spacing_val):
    """Input boxes where users can select the range and tick spacing for
    plot axes"""
    return dash.html.Div([
        dash.html.H5('Axis Range'),
        dash.html.P(
            ('Sets the minimum and/or maximum axis limits.  Leave blank to '
             'automatically scale the axis'),
            style={'marginBottom': UI_DESCRIPTION_MARGIN_BELOW},
        ),
        dash.html.Div([
            dbc.Input(
                debounce=True,
                id=range_min_id,
                value=range_min_val,
                style={'display': 'inline-block', 'width': '100px'},
            ),
            dash.html.P('  —  ', style={'display': 'inline-block',
                                        'whiteSpace': 'pre'}),
            dbc.Input(
                debounce=True,
                id=range_max_id,
                value=range_max_val,
                style={'display': 'inline-block', 'width': '100px'},
            ),
        ]),
        dash.html.Br(),
        dash.html.H5('Tick Spacing'),
        dash.html.P(
            ('Sets the distance between tick marks.  Leave blank to '
             'automatically scale the tick spacing'),
            style={'marginBottom': UI_DESCRIPTION_MARGIN_BELOW},
        ),
        dbc.Input(
            debounce=True,
            id=tick_spacing_id,
            value=tick_spacing_val,
            style={'width': '100px'},
        ),
    ])



"""This module provides a class intended for reading, parsing, and writing
files which store simulation results produced by the Maha Multics software.
"""

import copy
import pathlib
import re
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pyxx

from mahautils.utils.dictionary import Dictionary
from .configfile import MahaMulticsConfigFile
from .exceptions import (
    InvalidSimResultsFormatError,
    SimResultsError,
    SimResultsDataNotFoundError,
    SimResultsKeyError,
    SimResultsOverwriteError,
)


class _SimResultsEntry:
    """An object for storing information about simulation results variables

    This class is intended used by :py:class:`SimResults` to store metadata
    and (possibly) simulation results data for variables in Maha Multics
    simulation results files.
    """

    # Whether to include raw data when converting `_SimResultsEntry` objects
    # to a printable string representation
    show_data = True

    def __init__(self,
                 required: bool,
                 units: str,
                 data: Optional[
                     Union[np.ndarray, List[float], Tuple[float, ...]]] = None,
                 description: Optional[str] = None,
                 group: Optional[str] = None,
                 ) -> None:
        # Mypy exclusions in constructor were added as workarounds
        # for python/mypy#3004
        self.data = data  # type: ignore
        self.description = description
        self.group = group
        self.required = required
        self.units = units

    def __repr__(self) -> str:
        representation = (
            f'[{"Required" if self.required else "Optional"}] '
            f'[Units: {self.units}]'
        )

        if self.description is not None:
            representation += f' {self.description}'

            if (self.show_data) and (self.data is not None):
                representation += ','

        if (self.show_data) and (self.data is not None):
            array_str = np.array2string(self.data, precision=2)
            representation += f' {array_str}'

        return representation

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def data(self) -> Union[np.ndarray, None]:
        """A time series of simulation results data for the given variable"""
        return self._data

    @data.setter
    def data(self,
             data: Optional[Union[
                 np.ndarray, List[float], Tuple[float, ...]]]
             ) -> None:
        if data is None:
            self._data = None
        else:
            self._data = np.array(data, dtype=np.float64)

            if self._data.ndim != 1:
                raise ValueError('Simulation results data must be 1D')

    @property
    def description(self) -> Union[str, None]:
        """A string describing the variable"""
        return self._description

    @description.setter
    def description(self, description: Optional[str]) -> None:
        if description is None:
            self._description = None
        else:
            self._description = str(description)

    @property
    def group(self) -> Union[str, None]:
        """A string identifying a group to which multiple simulation results
        variables may belong"""
        return self._group

    @group.setter
    def group(self, group: Optional[str]) -> None:
        if group is None:
            self._group = None
        else:
            self._group = str(group)

    @property
    def required(self) -> bool:
        """Whether the Maha Multics software is required to output data for
        the simulation results variable when running

        If unable to output the variable, the Maha Multics software should
        exit with error.
        """
        return self._required

    @required.setter
    def required(self, required: bool) -> None:
        self._required = bool(required)

    @property
    def units(self) -> str:
        """The units of the data stored for the simulation results variable"""
        return self._units

    @units.setter
    def units(self, units: str) -> None:
        self._units = str(units)


class SimResults(MahaMulticsConfigFile):
    """An object representing a Maha Multics simulation results file

    The Maha Multics software uses simulation results files to configure which
    simulation results will be generated and to store the results.  This class
    is capable of parsing such files and providing an interface through which
    users can interact with and manipulate data in the file.
    """

    def __init__(self, path: Optional[Union[str, pathlib.Path]] = None,
                 unit_converter: Optional[pyxx.units.UnitConverter] = None
                 ) -> None:
        """Creates an object that can read and write Maha Multics simulation
        results files

        Creates an instance of the :py:class:`SimResults` class and optionally
        reads and parses a specified Maha Multics simulation results file.

        Parameters
        ----------
        path : str or pathlib.Path, optional
            The path and filename of the simulation results file to read and
            parse (default is ``None``).  If set to ``None``, no file is read
        unit_converter : pyxx.units.UnitConverter, optional
            A :py:class:`pyxx.units.UnitConverter` instance which will be
            used to convert units of quantities stored in the simulation
            results file (default is ``None``).  If set to ``None``, the
            :py:class:`MahaMulticsUnitConverter` unit converter will be used
            to perform unit conversions
        """
        super().__init__(path=path, unit_converter=unit_converter)

        self.__printdict_begin_regex = r'^\s*printDict\s*{\s*'
        self.__printdict_end_regex = r'^\s*}\s*$'

        self.__maha_multics_version_identifier = 'Multics Version'
        self.__maha_multics_commit_identifier = 'Multics Git Commit Hash'
        self.__sim_version_identifier = 'Main Sketch Version'

        # Initialize variables
        self._compile_options: Dict[str, str] = {}
        self._data: Dictionary[str, _SimResultsEntry] = Dictionary(
            custom_except_class=SimResultsKeyError,
            custom_except_msg='Variable "%s" not found in simulation results file'
        )
        self._maha_multics_commit: Union[str, None] = None
        self._maha_multics_version: Union[str, None] = None
        self._num_time_steps: int = 0
        self._sim_version: Union[str, None] = None
        self._title: Union[str, None] = None
        self.trailing_newline = True

        # If path was provided, read file
        if path is not None:
            self.read(path, parse=True)

    def __repr__(self) -> str:
        representation = [
            f'{self.__class__}',
            f'Title:      {self.title}',
        ]

        representation.append(f'Time steps: {self.num_time_steps}')

        representation.extend([
            f'Hashes:     {self.hashes}',
            '',
            self.__str__(),
        ])

        return '\n'.join(representation)

    def __str__(self) -> str:
        _SimResultsEntry.show_data = True
        return self.__get_printable_vars_str(self.variables)

    @property
    def compile_options(self) -> Dict[str, str]:
        """A dictionary containing options with which the Maha Multics
        software was compiled

        Note that this dictionary is returned **by reference**, so modifying
        the returned value will modify the object attribute.
        """
        return self._compile_options

    @property
    def maha_multics_commit(self) -> Union[str, None]:
        """The Git commit hash of the Maha Multics software with which the simulation
        was compiled"""
        return self._maha_multics_commit

    @maha_multics_commit.setter
    def maha_multics_commit(self, maha_multics_commit: Union[str, None]) -> None:
        if not isinstance(maha_multics_commit, (str, type(None))):
            raise TypeError('Argument "maha_multics_commit" must be of type '
                            '"str" or "None"')

        self._maha_multics_commit = maha_multics_commit

    @property
    def maha_multics_version(self) -> Union[str, None]:
        """The version of the Maha Multics software with which the simulation
        was compiled"""
        return self._maha_multics_version

    @maha_multics_version.setter
    def maha_multics_version(self, maha_multics_version: Union[str, None]) -> None:
        if not isinstance(maha_multics_version, (str, type(None))):
            raise TypeError('Argument "maha_multics_version" must be of type '
                            '"str" or "None"')

        self._maha_multics_version = maha_multics_version

    @property
    def num_time_steps(self) -> int:
        """The number of time steps in the data array of the simulation
        results file"""
        return self._num_time_steps

    @property
    def sim_version(self) -> Union[str, None]:
        """The simulation version of the ``main.cpp`` file which was used to
        generate the data in the simulation results file"""
        return self._sim_version

    @sim_version.setter
    def sim_version(self, sim_version: Union[str, None]) -> None:
        if not isinstance(sim_version, (str, type(None))):
            raise TypeError('Argument "sim_version" must be of type '
                            '"str" or "None"')

        self._sim_version = sim_version

    @property
    def title(self) -> Union[str, None]:
        """A title describing the contents of the simulation results file"""
        return self._title

    @title.setter
    def title(self, title: Union[str, None]) -> None:
        if not isinstance(title, (str, type(None))):
            raise TypeError('Argument "title" must be of type "str" or "None"')

        self._title = title

    @property
    def variables(self) -> Tuple[str, ...]:
        """A tuple containing all variable names listed in the simulation
        results file

        This property contains all variables listed in the "printDict" section
        of the simulation results file.  There may or may not be simulation
        results data available for any given variable.
        """
        return tuple(self._data.keys())

    def __get_printable_vars_str(self,
                                 variables: Union[List[str], Tuple[str, ...]],
                                 indent: int = 2
                                 ) -> str:
        if len(variables) == 0:
            representation = '[No simulation results variables found]'

        else:
            max_key_len = pyxx.arrays.max_list_item_len(variables)

            representation = ''

            for group in (list(self.list_groups()) + [None]):
                var_str = [group if group is not None else 'No Group Assigned']

                for var in variables:
                    if self._data[var].group == group:
                        var_str.append(
                            f'{" "*indent}{var:{max_key_len+1}s}: '
                            f'{str(self._data[var])}'
                        )

                if len(var_str) > 1:
                    representation += '\n'.join(var_str) + '\n'

        return representation.strip('\n')

    def __parse_metadata_comment(self, identifier: str) -> Union[str, None]:
        for line in self.contents:
            if re.match(r'^\s*#\s*' + re.escape(identifier) + r'\:', line):
                return line.split(f'{identifier}:', maxsplit=1)[1].strip()

        return None

    def _remove_vars_with_asterisk(self) -> None:
        # Extract names of all simulation results variables whose data are
        # included in the file
        data_vars: List[str] = []
        i = 0
        while i < len(self.contents):
            if (line := self.contents[i]).startswith('$'):
                data_vars = [var.split(':')[0] for var in line.split('$')[1:]]
                break

            i += 1

        i = 0
        while i < len(self.contents):
            line = self.contents[i]

            if (
                groups := re.search(self.__printdict_begin_regex + r'(.*)', line)
            ) is not None:
                if (line := groups.groups()[-1]) != '':
                    self.contents[i] = 'printDict{'
                    self.contents.insert(i + 1, '    ' + line)

                i += 1
                line = self.contents[i]

                while not re.match(self.__printdict_end_regex, line):
                    # Identify problematic keys (with asterisks)
                    line_groups = re.search(
                        r'^\s*([@?])\s*([\w\d\._\*\(\,\)]+)\s+\[([^\s]+)\]\s*$',
                        line)

                    # If key with asterisk was found, remove it and replace
                    # with simulation results variables
                    if line_groups is not None:
                        asterisk_key = line_groups.group(2)

                        if '*' in asterisk_key:
                            if len(data_vars) == 0:
                                raise SimResultsDataNotFoundError(
                                    'Simulation results files with asterisks (*) '
                                    'in variable names can only be read if the '
                                    'line beginning with "$" providing detailed '
                                    'variable names and descriptions is present')

                            # Find any matching simulation results variables
                            asterisk_seqs = (
                                re.findall(r'\*\(\d+\,\d+\)', asterisk_key)
                                + ['**']
                            )

                            key_regex = asterisk_key
                            for seq in asterisk_seqs:
                                key_regex = key_regex.replace(seq, '*')

                            key_regex = re.escape(key_regex).replace(r'\*', r'\d+')

                            matching_keys = []
                            for var in data_vars:
                                if re.match(key_regex, var):
                                    matching_keys.append(var)

                            matching_keys.reverse()

                            # Add matching simulation results variables to file
                            del self.contents[i]

                            for var in matching_keys:
                                self.contents.insert(
                                    i, line.replace(asterisk_key, var))

                            i += len(matching_keys) - 1

                    i += 1
                    line = self.contents[i]

                # Coverage.py is unable to detect 'break' statements
                # https://github.com/nedbat/coveragepy/issues/772
                break  # pragma: no cover

            i += 1

    def append(self, key: str, required: bool, units: str,
               data: Optional[
                   Union[np.ndarray, List[float], Tuple[float, ...]]] = None,
               description: Optional[str] = None,
               group: Optional[str] = None,
               ) -> None:
        """Adds a variable to the simulation results file

        Adds a new variable to the simulation results file, placing it at the
        end of the :py:attr:`variables` list.

        Parameters
        ----------
        key : str
            A unique name to identify the variable
        required : bool
            Whether the simulation results variable is required to be
            outputted by Maha Multics
        units : str
            The units of the variable's data
        data : np.ndarray or list or tuple, optional
            If desired, a 1D array consisting of the simulation results data
            for the variable can be provided (default is ``None``)
        description : str, optional
            An optional description of the simulation results variable
            (default is ``None``)
        group : str, optional
            An optional group to which the variable belongs (default is
            ``None``)

        Raises
        ------
        SimResultsOverwriteError
            If a variable ``key`` already exists in the simulation results file
        """
        if key in self.variables:
            raise SimResultsOverwriteError(
                f'Simulation results variable "{key}" already exists')

        self._data[key] = _SimResultsEntry(
            required    = required,
            units       = units,
            data        = None,
            description = description,
            group       = group
        )

        self.set_data(key, data, units)

    def clear_data(self, regex_pattern: str = '.+') -> List[str]:
        """Removes simulation results data for any variable(s) with names
        matching a given regex pattern

        Parameters
        ----------
        regex_pattern : str, optional
            Any variables in :py:attr:`variables` matching this regex pattern
            will have existing simulation results data deleted (default is
            ``'.+'``, which matches all variable names)

        Returns
        -------
        list
            A list containing the variable names whose data were deleted
        """
        cleared_vars = []

        for var in self.variables:
            if re.match(regex_pattern, var):
                cleared_vars.append(var)
                self._data[var].data = None

        keys_without_data = [x.data is None for x in self._data.values()]
        if all(keys_without_data):
            self._num_time_steps = 0

        return cleared_vars

    def get_data(self, key: str, units: str) -> np.ndarray:
        """Extracts simulation results data by variable name

        Extracts data corresponding to a specific simulation results variable
        and returns it as a NumPy array with user-specified units, performing
        a unit conversion if necessary.

        Parameters
        ----------
        key : str
            The name of the variable whose data to extract
        units : str
            The units in which the simulation results data should be returned

        Returns
        -------
        np.ndarray
            A NumPy array containing the simulation results data for variable
            ``key``, in units of ``units``
        """
        # Extract data
        data_ref = self._data[key].data

        if data_ref is None:
            raise SimResultsDataNotFoundError(
                f'No simulation results data are present for key "{key}"')

        data = data_ref.copy()
        stored_units = self._data[key].units

        # Perform unit conversion if necessary
        if units == stored_units:
            return data

        return self.unit_converter.convert(
            quantity=data, from_unit=stored_units, to_unit=units)

    def get_description(self, key: str) -> Union[str, None]:
        """Returns the description (if assigned) of a variable in the
        simulation results file

        Parameters
        ----------
        key : str
            The name of the variable whose description to retrieve

        Returns
        -------
        str or None
            A string containing the description of the simulation results
            variable, or ``None`` if no description is available
        """
        return self._data[key].description

    def get_group(self, key: str) -> Union[str, None]:
        """Returns the name of the group (if assigned) of a variable in the
        simulation results file

        Parameters
        ----------
        key : str
            The name of the variable whose group to retrieve

        Returns
        -------
        str or None
            A string containing the name of the group of the simulation
            results variable, or ``None`` if no group has been assigned
        """
        return self._data[key].group

    def get_required(self, key: str) -> bool:
        """Returns whether a variable in the simulation results file is
        specified as "required"

        Variables specified as "required" indicates to the Maha Multics
        software that the simulation should exit with error if unable to
        output these values.

        Parameters
        ----------
        key : str
            The name of the variable whose data to extract

        Returns
        -------
        bool
            Whether the variable specified by ``key`` is marked as "required"
            in the simulation results file
        """
        return self._data[key].required

    def get_units(self, key: str) -> str:
        """Returns the "raw" units in which data in the simulation results
        file are stored

        When extracting data with :py:meth:`get_data`, the units can be
        converted to any valid unit; however, the data in :py:class:`SimResults`
        objects is internally stored with a particular set of units.  This
        method returns these units of the internally stored data.

        Parameters
        ----------
        key : str
            The name of the variable whose units to return

        Returns
        -------
        str
            The internally stored units of data in the simulation results file
        """
        return self._data[key].units

    def list_groups(self) -> Tuple[str, ...]:
        """Returns a tuple containing all variable groups in the simulation
        results file

        Returns
        -------
        tuple
            A tuple of strings, where each string is a variable group in the
            simulation results file
        """
        groups = []
        for entry in self._data.values():
            if (entry.group is not None) and (entry.group not in groups):
                groups.append(entry.group)

        return tuple(groups)

    def parse(self) -> None:
        """Parses the file content in the :py:attr:`contents` list and
        populates attributes (such as :py:attr:`title`) with extracted data

        This method parses the data in :py:attr:`contents`, extracting
        configuration and simulation results data and storing it in this
        object's attributes for easier reading and editing.
        """
        # Verify that file contents have been read
        super().parse()

        original_contents = copy.deepcopy(self.contents)

        # Extract title
        self.title = self.__parse_metadata_comment('Title')
        self.maha_multics_version \
            = self.__parse_metadata_comment(self.__maha_multics_version_identifier)
        self.maha_multics_commit \
            = self.__parse_metadata_comment(self.__maha_multics_commit_identifier)
        self.sim_version \
            = self.__parse_metadata_comment(self.__sim_version_identifier)

        # Extract options with which Maha Multics software was compiled
        for line in self.contents:
            if line.strip().startswith('#_OPTIONs:'):
                options_list = line.split('#_OPTIONs:', maxsplit=1)[1].split(',')

                for opt in options_list:
                    opt_parsed = [x.strip() for x in opt.split('=', maxsplit=1)]
                    self._compile_options[opt_parsed[0]] = opt_parsed[1]

                break

        # Pre-process input file (to prepare to read simulation results data)
        self.clean_contents(
            remove_comments    = False,
            strip              = True,
            concat_lines       = False,
            remove_blank_lines = True
        )

        self._remove_vars_with_asterisk()

        # Extract list of variables in "printDict"
        sim_results_vars, comments, _, num_sec = self.extract_section_by_keyword(
            section_label      = 'printDict',
            begin_regex        = self.__printdict_begin_regex,
            end_regex          = self.__printdict_end_regex,
            section_line_regex = r'^\s*([@?])\s*([\w\d\._]+)\s+\[([^\s]+)\]\s*$',
            max_sections       = 1,
        )

        if num_sec == 0:
            raise InvalidSimResultsFormatError(
                'Unable to find "printDict" section in simulation results file')

        group_name = None
        for i, var in enumerate(sim_results_vars):
            if len(comments[i]) > 0:
                group_name = '\n'.join(
                    [self._extract_full_line_comment_text(x)
                     for x in comments[i]]
                )

            required_char, key, units = var.groups()

            if required_char == '@':
                required = True
            elif required_char == '?':
                required = False
            else:
                # This condition should never be reached (an error would have
                # been thrown when attempting to extract the section since the
                # regex wouldn't have matched) but it is included as a backup
                raise InvalidSimResultsFormatError(
                    f'Invalid character "{required_char}" indicating whether '
                    'a variable is required. The only valid characters are '
                    '"@" and "?"')  # pragma: no cover

            self._data[key] = _SimResultsEntry(required, units, group=group_name)

        # Extract simulation data
        self.clean_contents(
            remove_comments         = True,
            skip_full_line_comments = False,
            strip                   = True,
            concat_lines            = False,
            remove_blank_lines      = True
        )

        self._num_time_steps = 0

        i = 0
        while i < len(self.contents):
            line = self.contents[i]

            if line.startswith('$'):
                # Parse list of variables for which simulation data
                # are present
                data_array_vars: List[List[str]] = [
                    var.split(':') for var in line.split('$')[1:]
                ]

                for var_data in data_array_vars:
                    if len(var_data) != 3:
                        raise InvalidSimResultsFormatError(
                            'In data array variable list, all entries should '
                            'be colon-separated key, units, and description. '
                            'The following variable does not fit this format: '
                            f'{":".join(var_data)}'
                        )

                    key = var_data[0]
                    units = var_data[1]
                    description = var_data[2] if len(var_data[2]) > 0 else None

                    if key not in self._data:
                        raise InvalidSimResultsFormatError(
                            f'Key "{key}" was found in simulation results data '
                            'array but not in "printDict" section'
                        )

                    if self._data[key].units != units:
                        raise InvalidSimResultsFormatError(
                            f'For key "{key}," units in "printDict" section '
                            f'({self._data[key].units}) and in data array '
                            f'section ({units}) do not match'
                        )

                    self._data[key].description = description

                    num_data_array_vars = len(data_array_vars)

                # Parse simulation data
                try:
                    # Replace "nan(ind)" with "nan" since MSVC compiler outputs
                    # NaN as "nan(ind)" but this can't be converted by NumPy
                    data_array = np.transpose(np.array(
                        [x.replace('nan(ind)', 'nan').split()
                         for x in self.contents[i+1:]],
                        dtype=np.float64,
                    ))

                    if data_array.ndim != 2:
                        # Due to the list comprehension above, this condition
                        # should never occur; however, it is checked anyway to
                        # be safe
                        raise AssertionError(  # pragma: no cover
                            'Simulation results data array must be 2D')

                    if data_array.shape[0] != num_data_array_vars:
                        raise AssertionError(
                            f'Expected {num_data_array_vars} in simulation '
                            'results data array, but found '
                            f'{data_array.shape[0]}'
                        )

                except (AssertionError, ValueError) as exception:
                    raise InvalidSimResultsFormatError(
                        'Unable to read simulation results data array'
                    ) from exception

                for i in range(num_data_array_vars):
                    key = data_array_vars[i][0]
                    self._data[key].data = data_array[i]

                # Store number of time steps of simulation results
                self._num_time_steps = data_array.shape[1]

                break

            i += 1

        self.contents.clear()
        self.contents.extend(original_contents)

    def remove(self, key: str) -> None:
        """Removes a simulation results variable

        Removes a variable and any associated data from the simulation results
        file.

        Parameters
        ----------
        key : str
            The name of the variable to remove
        """
        if key not in self.variables:
            raise SimResultsKeyError(
                f'Simulation results variable "{key}" does not exist')

        del self._data[key]

    def search(self, keyword: str,
               search_fields: Union[
                   str, Tuple[str, ...], List[str]
               ] = ('keys', 'description', 'group', 'units'),
               case_sensitive: bool = False,
               show_data: bool = False,
               print_results: bool = True,
               return_results: bool = False,
               ) -> Union[Tuple[str, ...], None]:
        """Searches the simulation results file for a given search term

        This method can be useful for finding data, particularly in a large
        simulation results file.  It performs a relatively basic search,
        checking to see if a given term ``keyword`` is in any of the specified
        search fields.  The exact ``keyword`` must be found to register a
        match; similar terms are not considered a match.  To match any string,
        set ``keyword`` to ``''``.

        Parameters
        ----------
        keyword : str
            The term to search for
        search_fields : str or tuple or list, optional
            One or more simulation results variable attributes in which to
            search for ``keyword`` (default is to search all possible
            attributes, ``('keys', 'description', 'group', 'units')``)
        case_sensitive : bool, optional
            Whether to perform a case-sensitive search (default is ``False``)
        show_data : bool, optional
            Whether to print simulation results data, if available, for each
            variable (default is ``False``).  Setting this to ``True`` can
            considerably increase output verbosity
        print_results : bool, optional
            Whether to display results to the terminal (default is ``True``)
        return_results : bool, optional
            Whether to return a tuple of strings containing the simulation
            results variable keys for all search matches (default is ``False``)

        Returns
        -------
        tuple
            A tuple of strings containing all simulation results variable keys
            (i.e., selected from :py:attr:`variables`) for which search
            matches were registered.  Only returned if ``return_results`` is
            ``True``
        """
        keyword = str(keyword)
        if not case_sensitive:
            keyword = keyword.lower()

        search_fields = pyxx.arrays.convert_to_tuple(search_fields)
        invalid_fields = (set(search_fields)
                          - set(('keys', 'description', 'group', 'units')))
        if len(invalid_fields) > 0:
            raise ValueError(f'Invalid search fields: {invalid_fields}')

        _SimResultsEntry.show_data = show_data

        matches = []
        for var in self.variables:
            if 'keys' in search_fields:
                key = var
                if not case_sensitive:
                    key = key.lower()

                if keyword in key:
                    matches.append(var)
                    continue

            if 'description' in search_fields:
                if (description := self._data[var].description) is not None:
                    if not case_sensitive:
                        description = description.lower()

                    if keyword in description:
                        matches.append(var)
                        continue

            if 'group' in search_fields:
                if (group := self._data[var].group) is not None:
                    if not case_sensitive:
                        group = group.lower()

                    if keyword in group:
                        matches.append(var)
                        continue

            if 'units' in search_fields:
                units = self._data[var].units
                if not case_sensitive:
                    units = units.lower()

                if keyword in units:
                    matches.append(var)
                    continue

        if print_results:
            print(self.__get_printable_vars_str(matches))

        if return_results:
            return tuple(matches)

        return None

    def set_data(self, key: str,
                 data: Union[np.ndarray, List[float], Tuple[float, ...], None],
                 units: Optional[str] = None) -> None:
        """Adds or replaces simulation results data by variable name

        Parameters
        ----------
        key : str
            The name of the variable whose data to store
        data : np.ndarray or list or tuple
            A 1D array containing the data (in units of ``units``) to store
        units : str
            The units in which the simulation results data should be stored
        """
        if data is not None:
            # Store units
            if units is None:
                raise TypeError('If argument "data" is not `None`, argument '
                                '"units" must be provided')

            self._data[key].units = units

            # Validate data shape
            if self.num_time_steps == 0:
                self._num_time_steps = len(data)

            elif (self.num_time_steps > 0) and (len(data) != self.num_time_steps):
                raise ValueError(
                    f'Simulation results file has data for {self.num_time_steps} '
                    f'time steps. Cannot add new data series with {len(data)} '
                    'time steps')

        self._data[key].data = data

    def set_description(self, key: str, description: Union[str, None]) -> None:
        """Adds or modifies the description of a variable in the simulation
        results file

        Parameters
        ----------
        key : str
            The name of the variable whose description to retrieve
        description : str or None
            The description of the simulation results variable, or ``None`` to
            remove the description
        """
        self._data[key].description = description

    def set_group(self, key: str, group: Union[str, None]) -> None:
        """Sets or modifies the name of the group of a variable in the
        simulation results file

        Parameters
        ----------
        key : str
            The name of the variable whose group to retrieve
        description : str or None
            The group name of the simulation results variable, or ``None`` to
            remove the group name
        """
        self._data[key].group = group

    def set_required(self, key: str, required: bool) -> None:
        """Modifies whether a variable in the simulation results file is
        specified as "required"

        Variables specified as "required" indicates to the Maha Multics
        software that the simulation should exit with error if unable to
        output these values.

        Parameters
        ----------
        key : str
            The name of the variable whose data to extract
        required : bool
            Whether the variable specified by ``key`` is marked as "required"
            in the simulation results file
        """
        self._data[key].required = required

    def set_units(self, key: str, units: str,
                  action_if_data_present: str = 'error') -> None:
        """Modifies the units of a variable in the simulation results file

        Parameters
        ----------
        key : str
            The name of the variable whose units to modify
        units : str
            The units to assign to variable ``key``
        action_if_data_present : str, optional
            The action that should be taken if simulation results data for
            ``key`` are already present in the simulation results file
            (default is ``'error'``).  See "Notes" section for additional
            information

        Notes
        -----
        In the case that simulation results data for ``key`` already exist, it
        is not obvious how this case should be handled: should the data also
        be converted to the new units, or should only the units be changed?

        MahaUtils provides three options to handle this case:

        1. ``'error'``: If simulation results data are present, throw an error
        2. ``'convert_data'``: If simulation results data are present, convert
           it to the units given by the ``units`` argument (e.g., :math:`3\ m`
           would become :math:`0.003\ mm`)
        3. ``'keep_data_values'``: If simulation results data are present,
           keep the same values of the data but change the units (e.g.,
           :math:`3\ m` would become :math:`3\ mm`)
        """
        # Validate inputs
        valid_actions = ['error', 'convert_data', 'keep_data_values']
        if action_if_data_present not in valid_actions:
            raise ValueError('Action if data present must be selected '
                             f'from {valid_actions}')

        # Take no action if units match existing units
        if self._data[key].units == units:
            return

        # Store units
        if self._data[key].data is not None:
            if action_if_data_present == 'error':
                raise SimResultsOverwriteError(
                    f'Unable to set units for "{key}" because simulation '
                    'results data are already present')

            if action_if_data_present == 'convert_data':
                self._data[key].data = self.unit_converter.convert(
                    self._data[key].data,
                    from_unit=self._data[key].units, to_unit=units
                )

            if action_if_data_present == 'keep_data_values':
                pass

        self._data[key].units = units

    def update_contents(self, add_sim_data: bool = True, indent: int = 4,
                        padding: int = 4) -> None:
        """Updates the :py:attr:`contents` list based on object attributes

        This method synchronizes the :py:attr:`contents` list with the data
        stored in the :py:class:`SimResults` attributes (:py:attr:`title`,
        variables added by :py:meth:`append`, etc.).  This step should
        generally be performed prior to calling :py:meth:`write`, as writing
        a file directly saves the data in :py:attr:`contents` to disk.

        Parameters
        ----------
        add_sim_data : bool, optional
            Whether to add simulation results data to the end of
            :py:attr:`contents` (default is ``True``)
        indent : int, optional
            The number of spaces to use for indentation in the file (default
            is ``4``)
        padding : int, optional
            The number of spaces to place between simulation results variable
            names and the units in the ``printDict`` section of the simulation
            results file (default is ``4``)
        """
        if self.comment_chars is None:
            # This condition should not be reached since comment characters
            # are hard-coded for simulation results files, but the check is
            # still present to be safe
            raise SimResultsError(  # pragma: no cover
                'Cannot update contents until the "comment_chars" attribute '
                'is defined')
        comment_char = self.comment_chars[0]

        self.contents.clear()

        # Write title
        if self.title is not None:
            self.contents.append(f'{comment_char} Title: {self.title}')
            self.contents.append('')

        # Write "printDict" section
        self.contents.append('printDict{')

        if len(self.variables) > 0:
            max_key_len = pyxx.arrays.max_list_item_len(self.variables)
            for group in ([None] + list(self.list_groups())):
                group_lines = []
                group_vars_added = False

                if group is not None:
                    group_lines.append(f'{" "*indent}{comment_char} {group}')

                for var in self.variables:
                    if self._data[var].group == group:
                        group_vars_added = True
                        group_lines.append(
                            f'{" "*indent}'
                            f'{"@" if self._data[var].required else "?"}'
                            f'{var:{max_key_len}s}'
                            f'{" "*padding}[{self._data[var].units}]'
                        )

                # This should always be true, since the list of groups is
                # identified by the contents of "_data" (which guarantees
                # that each group should have at least one variable)
                if group_vars_added:  # pragma: no cover
                    self.contents.extend(group_lines + [''])
        else:
            self.contents.append('')

        self.contents[-1] = '}'

        # Write additional metadata
        if self.maha_multics_version is not None:
            self.contents.append(f'# {self.__maha_multics_version_identifier}: '
                                 f'{self.maha_multics_version}')

        if self.maha_multics_commit is not None:
            self.contents.append(f'# {self.__maha_multics_commit_identifier}: '
                                 f'{self.maha_multics_commit}')

        if self.sim_version is not None:
            self.contents.append(f'# {self.__sim_version_identifier}: '
                                 f'{self.sim_version}')

        # Write simulation data, if available
        if add_sim_data:
            # Simulation results variables and descriptions
            data_vars = ''
            data_array = []
            for var in self.variables:
                data = self._data[var].data

                if data is not None:
                    if (description := self._data[var].description) is None:
                        description = ''

                    data_vars += (f'${var}:{self._data[var].units}:{description}')

                    data_array.append(data)

            # Simulation results data
            if len(data_vars) > 0:
                self.contents.append(data_vars)

                # Compilation options
                if len(self.compile_options) > 0:
                    compile_opts_str = ', '.join(
                        [f'{k} = {v}' for k, v in self.compile_options.items()])
                    self.contents.append(f'#_OPTIONs: {compile_opts_str}'.strip())

                # Simulation results data
                for line in np.transpose(np.array(data_array, dtype=np.float64)):
                    self.contents.append(' '.join([str(x) for x in line]))



"""Classes for representing a point in 3D space

The classes in this module can be used to store the location of a point in
a three-dimensional space.
"""

import math
from typing import Optional, Union

from .point import Array_Float3, Point


class CartesianPoint3D(Point):
    """Class representing a point in 3D Cartesian coordinates

    This class can be used to represent a point in the 3D Cartesian coordinate
    system.  Note that although the axes for such a coordinate system can be
    arbitrary (:math:`xyz`-coordinates, :math:`a_1 a_2 a_3`-coordinates, etc.),
    in this class the axes are always referred to as :math:`x`, :math:`y`, and
    :math:`z` for simplicity and clarity.

    Notes
    -----
    The equality operator (``==``) is defined for points.  Points are
    considered equal if they are of the same type (:py:class:`Point`,
    :py:class:`CartesianPoint3D`, etc.), have :py:attr:`coordinates`
    attributes of the same shape and values, and have the same value of
    :py:attr:`units`.

    Examples
    --------
    Create a :py:class:`CartesianPoint3D` with no location initialized:

    >>> print(mahautils.shapes.CartesianPoint3D())
    ()

    Create a :py:class:`CartesianPoint3D` with location specified by
    positional arguments:

    >>> print(mahautils.shapes.CartesianPoint3D(1, 2.3, 4))
    (1.0, 2.3, 4.0)
    >>> print(mahautils.shapes.CartesianPoint3D([4, 5, 6]))
    (4.0, 5.0, 6.0)
    >>> pnt = mahautils.shapes.CartesianPoint3D([6, 7, 8])
    >>> print(mahautils.shapes.CartesianPoint3D(pnt))
    (6.0, 7.0, 8.0)

    Create a :py:class:`CartesianPoint3D` with location specified by
    keyword arguments:

    >>> print(mahautils.shapes.CartesianPoint3D(x=1, y=2, z=3))
    (1.0, 2.0, 3.0)
    """

    def __init__(self, *args: Union[Array_Float3, 'CartesianPoint3D', float],
                 units: Optional[str] = None, **kwargs):
        """Defines a point in the 3D Cartesian coordinate system

        Creates a :py:class:`CartesianPoint3D` instance and optionally allows
        the user to define the location of the point.

        Parameters
        ----------
        args : list or tuple or CartesianPoint3D or float, optional
            Positional arguments provided when creating the point object.  See
            the "Notes" section for information on how to use positional
            arguments to specify the point location
        units : str, optional
            The units in which the geometry is defined, or ``None`` to
            indicate dimensionless geometry or that units are to be ignored
            (default is ``None``)
        kwargs : Any, optional
            Keyword arguments provided when creating the point object.  See
            the "Notes" section for information on how to use keyword arguments
            to specify the point location

        Notes
        -----
        When creating a :py:class:`CartesianPoint3D` instance, it is possible
        to provide the point location using the constructor arguments.  This
        location can be provided using *either positional or keyword arguments,
        but not both*.

        If providing the location using positional arguments, then any of the
        following may be specified: (1) three floating-point numbers; (2) a list,
        tuple, NumPy array, or any other array-like object containing two
        floating-point numbers; or (3) another :py:class:`CartesianPoint3D`
        instance.

        If providing the location using keyword arguments, then *three* keyword
        arguments *must* be specified: ``x``, ``y``, and ``z``.  Both arguments
        must be numeric types (integer or floating-point values).
        """
        super().__init__(units=units)

        # Variable that indicates whether the point coordinates have already
        # been stored
        stored_point = False

        # Store point location passed as positional arguments
        if len(args) > 0:
            # If user passed a list, tuple, or `CartesianPoint3D` instance,
            # directly store the point coordinates
            if len(args) == 1:
                self.coordinates = args[0]

            # If user passed individual point values, store them
            else:
                self.coordinates = args

            # Record that point location was stored
            stored_point = True

        # Store point location passed as keyword arguments
        if ('x' in kwargs) or ('y' in kwargs) or ('z' in kwargs):
            # Check that the user didn't try to provide the point coordinates
            # using both positional and keyword arguments
            if stored_point:
                raise TypeError(
                    'Invalid input. Point location can be specified using '
                    'either positional or keyword arguments, but not both')

            # Verify that both x-, y-, and z-coordinates were provided
            if not (('x' in kwargs) and ('y' in kwargs) and ('z' in kwargs)):
                raise TypeError(
                    'Invalid keyword arguments. If providing point location '
                    'using keyword arguments, both arguments "x," "y," and '
                    '"z" must be provided')

            # Store point coordinates
            self.coordinates = (kwargs['x'], kwargs['y'], kwargs['z'])

            # Record that point location was stored
            stored_point = True

    @property
    def coordinates(self):
        """The coordinates of the point, represented as a tuple ``(x, y, z)``"""
        return super().coordinates

    @coordinates.setter
    def coordinates(self, coordinates: Union[Array_Float3, 'CartesianPoint3D']):
        if isinstance(coordinates, CartesianPoint3D):
            self._coordinates = coordinates.coordinates

        else:
            # Verify that three coordinates were provided
            try:
                # This does not use `assert` because `assert` statements can in
                # some cases be removed when compiling to optimized byte code
                if len(coordinates) != 3:
                    raise AssertionError(
                        'Length of provided coordinates must be 3, '
                        f'but {len(coordinates)} coordinate(s) were provided')
            except (AssertionError, TypeError) as exception:
                raise ValueError('Point coordinates must be a list or tuple '
                                 'of length 3') from exception

            # Convert coordinates to floating-point numbers and store values
            try:
                self._coordinates \
                    = (float(coordinates[0]), float(coordinates[1]),
                       float(coordinates[2]))
            except (TypeError, ValueError) as exception:
                raise ValueError(
                    'All point coordinates must be of a numeric type '
                    '(float, int, etc.)') from exception

    @property
    def x(self):
        """The x-coordinate of the point"""
        return self.coordinates[0]

    @x.setter
    def x(self, x: float) -> None:
        self.coordinates = (x, self.y, self.z)

    @property
    def y(self):
        """The y-coordinate of the point"""
        return self.coordinates[1]

    @y.setter
    def y(self, y: float) -> None:
        self.coordinates = (self.x, y, self.z)

    @property
    def z(self):
        """The z-coordinate of the point"""
        return self.coordinates[2]

    @z.setter
    def z(self, z: float) -> None:
        self.coordinates = (self.x, self.y, z)

    def distance_to(self, point: Union[Array_Float3, 'CartesianPoint3D']):
        """Computes the distance to another point

        Calculates and returns the distance to another point in the same 3D
        Cartesian coordinate system.

        Parameters
        ----------
        point : list or tuple or CartesianPoint3D
            The point to which to calculate distance

        Returns
        -------
        float
            The distance to another location ``point`` in 3D Cartesian space
        """
        pnt = CartesianPoint3D(point)
        return math.sqrt((self.x - pnt.x)**2 + (self.y - pnt.y)**2
                         + (self.z - pnt.z)**2)



"""This module provides a class intended for reading, parsing, and writing
files which store polygon geometry used by the Maha Multics software.
"""

import copy
import math
import pathlib
import re
from typing import List, Optional, Union

# Mypy type checking disabled for packages that are not PEP 561-compliant
import numpy as np
import plotly.express as px        # type: ignore
import plotly.graph_objects as go  # type: ignore
import pyxx

from mahautils.shapes.geometry.polygon import Polygon
from mahautils.shapes.geometry.shape_open_closed import ClosedShape2D
from mahautils.shapes.plotting import _create_blank_plotly_figure, _figure_config
from mahautils.shapes.layer import Layer
from mahautils.utils.dictionary import Dictionary
from .configfile import MahaMulticsConfigFile
from .exceptions import PolygonFileFormatError, PolygonFileMissingDataError


class PolygonFile(MahaMulticsConfigFile):
    """An object representing a Maha Multics polygon file

    The Maha Multics software uses polygon files to store geometry used for
    setting boundary conditions when running fluid simulations.  This class
    is capable of parsing such files and providing an interface through which
    users can interact with and manipulate data in the file (through the
    :py:attr:`polygon_data` attribute).

    .. note::

        The data structure in :py:class:`PolygonFile` objects largely mirrors
        the content of Maha Multics polygon files, although there are a few
        differences (for which descriptive error messages are displayed).  For
        more information about the format of these files, refer to the
        :ref:`fileref-polygon_file` page.
    """

    def __init__(self, path: Optional[Union[str, pathlib.Path]] = None,
                 unit_converter: Optional[pyxx.units.UnitConverter] = None
                 ) -> None:
        super().__init__(path=path, unit_converter=unit_converter)

        # Initialize variables
        self._polygon_data: Dictionary[float, Layer] = Dictionary(
            required_key_type=(float, int),
            required_value_type=Layer,
        )
        self.trailing_newline = True

        self._polygon_merge_method: Union[int, None] = None
        self._time_extrap_method: Union[int, None] = None
        self._time_units: Union[str, None] = None

        # If path was provided, read file
        if path is not None:
            self.read(path, parse=True)

    @property
    def num_time_steps(self) -> int:
        """The number of time steps in the polygon file"""
        return len(self._polygon_data)

    @property
    def polygon_data(self) -> Dictionary[float, Layer]:
        """A reference to the dictionary containing the polygon data in the
        file

        This method returns a :py:class:`mahautils.utils.Dictionary` instance
        containing the polygon data in the file.  The keys are the time values
        defined in the file (with units given by :py:attr:`time_units`) and
        the values are :py:class:`Layer` instances containing the polygons
        corresponding to each time step.  This dictionary can be directly
        edited to modify the polygon file (adding new time steps, removing
        time steps, changing polygon properties, etc.).

        .. important::

            The dictionary is returned **by reference**, so editing the
            returned dictionary will directly edit data in the
            :py:class:`PolygonFile` instance.

        Notes
        -----
        This property can only be accessed if :py:attr:`polygon_merge_method`,
        :py:attr:`time_extrap_method`, and :py:attr:`time_units` are all
        defined (even for polygon files with a single time step, since
        additional time steps may be added by modifying the polygon data
        dictionary).  This ensures consistency of the data stored in the
        :py:class:`PolygonFile` object.  To access the polygon data before
        setting these properties, use :py:attr:`polygon_data_readonly`.

        Any of the methods for adding or removing data provided by
        :py:class:`mahautils.utils.Dictionary` can be used to modify polygon
        data.  For instance, to insert a new time step prior to an existing
        time, simply use :py:meth:`mahautils.utils.Dictionary.insert_before`.
        *This is why there is no "add polygon" method -- accessing the
        dictionary methods provides greater flexibility.*
        """
        # Verify that polygon file metadata is defined; otherwise it is
        # ambiguous what units and meaning user-specified polygons and time
        # steps have (i.e., if you add a new time step to the polygon data
        # dictionary at t=2, what units are t=2?)
        if None in [
            self._polygon_merge_method,
            self._time_extrap_method,
            self._time_units,
        ]:
            raise PolygonFileMissingDataError(
                'In order to edit or view the "polygon_data" dictionary, the '
                'following attributes must be set: (\'polygon_merge_method\', '
                '\'time_extrap_method\', \'time_units\')')

        return self._polygon_data

    @property
    def polygon_data_readonly(self) -> Dictionary[float, Layer]:
        """A copy of the dictionary containing the polygon data in the file

        This method returns a :py:class:`mahautils.utils.Dictionary` instance
        containing the polygon data in the file.  The keys are the time values
        defined in the file (with units given by :py:attr:`time_units`) and
        the values are :py:class:`Layer` instances containing the polygons
        corresponding to each time step.

        .. important::

            The dictionary is returned **by copy**, so editing the
            returned dictionary will **not** directly edit data in the
            :py:class:`PolygonFile` instance.

        Notes
        -----
        Unlike :py:attr:`polygon_data`, this method does **not** require the
        :py:attr:`polygon_merge_method`, :py:attr:`time_extrap_method`, and
        :py:attr:`time_units` properties to be defined defined.
        """
        return copy.deepcopy(self._polygon_data)

    @property
    def polygon_merge_method(self) -> int:
        """The method used to combine disjoint polygons

        This attribute represents the method for combining multiple disjoint
        polygons and is applied for all time steps (a limitation hard-coded
        in the Maha Multics software).

        For more information and a list of permissible values for the,
        :py:attr:`polygon_merge_method` property refer to the
        :ref:`Polygon File Format <fileref-polygon_file-polygon_merge>` page.
        """
        if self._polygon_merge_method is None:
            raise PolygonFileMissingDataError(
                'Attribute "polygon_merge_method" has not been defined')

        return self._polygon_merge_method

    @polygon_merge_method.setter
    def polygon_merge_method(self, polygon_merge_method: int) -> None:
        valid_choices = (0, 1, 2)
        if polygon_merge_method not in valid_choices:
            raise ValueError(
                'Polygon merge method (for combining multiple disjoint '
                f'polygons) must be one of the following: {valid_choices}'
            )

        self._polygon_merge_method = int(polygon_merge_method)

    @property
    def time_extrap_method(self) -> int:
        """The method used to handle time values outside the defined time range

        This property defines how the Maha Multics software will handle time
        values outside the range given by :py:attr:`time_values`.

        For more information and a list of permissible values for the,
        :py:attr:`time_extrap_method` property refer to the
        :ref:`Polygon File Format <fileref-polygon_file-time_extrap>` page.
        """
        if self._time_extrap_method is None:
            raise PolygonFileMissingDataError(
                'Attribute "time_extrap_method" has not been defined')

        return self._time_extrap_method

    @time_extrap_method.setter
    def time_extrap_method(self, time_extrap_method: int) -> None:
        valid_choices = (0, 2, 3)
        if time_extrap_method not in valid_choices:
            raise ValueError(
                'Time step extrapolation method must be one of the '
                f'following: {valid_choices}'
            )

        self._time_extrap_method = int(time_extrap_method)

    @property
    def time_units(self) -> str:
        """The units in which "time" values are stored in the polygon file

        Note that the term "time" is used loosely for polygon files, and "time"
        may also be given in terms of quantities such as shaft rotation angle.
        For more information, refer to the :ref:`fileref-polygon_file` page.
        """
        if self._time_units is None:
            raise PolygonFileMissingDataError(
                'Attribute "time_units" has not been defined')

        return self._time_units

    @time_units.setter
    def time_units(self, time_units: str) -> None:
        time_units = str(time_units)

        if not self.unit_converter.is_defined_unit(time_units):
            raise ValueError(f'Time units "{time_units}" are not defined in '
                             'this object\'s unit converter')

        self._time_units = time_units

    @property
    def time_values(self) -> List[float]:
        """The values of "time" defined in the polygon file

        This method returns a **copy** of the list of time values, so modifying
        the returned list will **not** affect the time values stored in the
        polygon file.  The time values are returned with units given by
        :py:attr:`time_units`.
        """
        return list(self._polygon_data.keys())

    def __parse_int(self, value: Union[str, float], error_message: str) -> int:
        if pyxx.numbers.is_integer(value):
            return int(float(value))

        raise PolygonFileFormatError(error_message)

    def filter_times(self, interval: float, tolerance: float = 1e-12) -> None:
        """Reduces the number of time steps in a polygon file

        Prunes time steps from a polygon file at user-specified intervals.
        This can be useful, for instance, if you have a polygon file with a
        time step of 0.01 seconds and you want to increase the time step to
        0.1 seconds.

        Parameters
        ----------
        interval : float
            The spacing at which to keep time steps, in units of
            :py:attr:`time_units`
        tolerance : float, optional
            The allowable difference between retained time steps and intervals
            of size ``interval`` (default is ``1e-12``)

        Notes
        -----
        Time steps are kept if they fall within a distance of ``tolerance``
        of ``{..., -2*interval, -interval, 0, interval, 2*interval, ...}``
        """
        for t in self.time_values:
            if abs(interval*round(t / interval) - t) > tolerance:
                del self._polygon_data[t]

    def parse(self) -> None:
        """Parses the file content in the :py:attr:`contents` list and
        populates attributes (such as the dictionary returned by
        :py:attr:`polygon_data`) with extracted data

        This method parses the data in :py:attr:`contents`, checking for
        polygon file format errors and extracting data on individual polygons
        for easier reading and modification.
        """
        super().parse()
        original_contents = copy.deepcopy(self.contents)

        # Remove comments and whitespace
        self.clean_contents(
            remove_comments    = True,
            strip              = True,
            concat_lines       = False,
            remove_blank_lines = False,
        )

        # Extract header
        line_idx = 0
        metadata = [x.strip() for x in self.contents[line_idx].split(maxsplit=2)]

        try:
            num_time_steps = self.__parse_int(
                value=metadata[0],
                error_message=(f'Error parsing polygon file (line {line_idx+1}): '
                               'number of time steps must be an integer'))
            polygons_per_time_step = self.__parse_int(
                value=metadata[1],
                error_message=(f'Error parsing polygon file (line {line_idx+1}): '
                               'polygons per time step must be an integer'))
            self.polygon_merge_method = self.__parse_int(
                value=metadata[2],
                error_message=(f'Error parsing polygon file (line {line_idx+1}): '
                               'disjoint polygon merge method must be an integer'))
        except IndexError as exception:
            raise PolygonFileFormatError(
                f'Error parsing polygon file (line {line_idx+1}): missing '
                'header data. The number of time steps, polygons per time '
                'step, and disjoint polygon merge method must be provided as '
                'whitespace-separated numbers'
            ) from exception
        except ValueError as exception:
            exception.args = ('Error parsing polygon file (line '
                              f'{line_idx+1}): {exception.args[0]}',)
            raise

        if num_time_steps > 1:
            line_idx += 1
            line = self.contents[line_idx]

            regex = r'^\s*(.+):\s*([\d.e+-]+)\s+([\d.e+-]+)\s+([\d.e+-]+)\s*$'
            groups = re.search(regex, line)

            if groups is None:
                raise PolygonFileFormatError(
                    f'Error parsing polygon file (line {line_idx+1}): time '
                    'values are not formatted correctly')

            try:
                self.time_units = str(groups.group(1))
            except ValueError as exception:
                exception.args = ('Error parsing polygon file (line '
                                  f'{line_idx+1}): {exception.args[0]}',)
                raise

            try:
                initial_time = float(groups.group(2))
                time_step = float(groups.group(3))
            except (TypeError, ValueError) as exception:
                raise PolygonFileFormatError(
                    f'Error parsing polygon file (line {line_idx+1}): initial '
                    'time and time step must be numbers'
                ) from exception

            try:
                self.time_extrap_method = self.__parse_int(
                    value=groups.group(4),
                    error_message=('Error parsing polygon file (line '
                                   f'{line_idx+1}): time step extrapolation '
                                   'method must be an integer')
                )
            except ValueError as exception:
                exception.args = ('Error parsing polygon file (line '
                                  f'{line_idx+1}): {exception.args[0]}',)
                raise

            time_vals = list(np.linspace(
                initial_time, time_step*(num_time_steps - 1) + initial_time,
                num_time_steps))

        else:
            time_vals = [0]

        # Extract polygon coordinates
        for i in range(num_time_steps):
            layer = Layer(print_multiline=False,
                          color=px.colors.qualitative.Plotly[0])

            for _ in range(polygons_per_time_step):
                if len(self.contents) < (line_idx + 2):
                    raise PolygonFileFormatError(
                        'Polygon file appears to be missing data for '
                        'some polygons')

                # Extract polygon metadata
                line_idx += 1

                try:
                    num_coordinates, enclosed_convention = (
                        [self.__parse_int(x,
                                          'Error parsing polygon file (line '
                                          f'{line_idx+1}): number of coordinates '
                                          'and polygon file enclosed convention '
                                          'must be integers')
                         for x in self.contents[line_idx].split(maxsplit=1)]
                    )
                except ValueError as exception:
                    raise PolygonFileFormatError(
                        f'Error parsing polygon file (line {line_idx+1}): '
                        'both number of coordinates and polygon file enclosed '
                        'convention must be provided'
                    ) from exception

                # Extract polygon x-coordinates
                line_idx += 1

                x_units, x_coordinates_str \
                    = self.contents[line_idx].split(':', maxsplit=1)
                x_units = x_units.strip()

                try:
                    x_coordinates = np.array(x_coordinates_str.split(),
                                             dtype=np.float64)
                except (TypeError, ValueError) as exception:
                    raise PolygonFileFormatError(
                        f'Error parsing polygon file (line {line_idx+1}): '
                        'x-coordinates are not a whitespace-separated list '
                        'of numbers'
                    ) from exception

                # Extract polygon y-coordinates
                line_idx += 1

                y_units, y_coordinates_str \
                    = self.contents[line_idx].split(':', maxsplit=1)
                y_units = y_units.strip()

                try:
                    y_coordinates = np.array(y_coordinates_str.split(),
                                             dtype=np.float64)
                except (TypeError, ValueError) as exception:
                    raise PolygonFileFormatError(
                        f'Error parsing polygon file (line {line_idx+1}): '
                        'y-coordinates are not a whitespace-separated list '
                        'of numbers'
                    ) from exception

                # Validate polygon data formatting
                if not (len(x_coordinates) == len(y_coordinates)
                        == num_coordinates):
                    raise PolygonFileFormatError(
                        f'Error parsing polygon file (lines {line_idx}-'
                        f'{line_idx+1}): expected {num_coordinates} point '
                        f'coordinates, but found {len(x_coordinates)} x-'
                        f'coordinates and {len(y_coordinates)} y-coordinates')

                for k, units in enumerate((x_units, y_units)):
                    if not self.unit_converter.is_defined_unit(units):
                        raise PolygonFileFormatError(
                            f'Error parsing polygon file (line {line_idx+k}): '
                            f'unrecognized units "{units}"')

                if x_units != y_units:
                    raise PolygonFileFormatError(
                        f'Error parsing polygon file (lines {line_idx}-'
                        f'{line_idx+1}): units for x-coordinates and y-'
                        'coordinates must be equal')

                # Create polygon from coordinates
                coordinates = (x_coordinates, y_coordinates)

                layer.append(
                    Polygon(vertices=np.transpose(np.stack(coordinates)),
                            polygon_file_enclosed_conv=enclosed_convention,
                            units=x_units)
                )

            self._polygon_data[time_vals[i]] = layer

        if line_idx != (len(self.contents) - 1):
            raise PolygonFileFormatError(
                'Polygon file contains more polygons than expected')

        self.contents.clear()
        self.contents.extend(original_contents)

    def plot(self, delay: float = 100, show: bool = True,
             return_fig: bool = False) -> Union[go.Figure, None]:
        """Generates an animated plot showing the geometry in the polygon file

        Polygons in the file are illustrated as filled, solid shapes, and
        construction geometry is illustrated as dotted outlines.

        Parameters
        ----------
        delay : float, optional
            The delay (in milliseconds) between each frame when animating the
            plot (default is ``100``)
        show : bool, optional
            Whether to open the figure in a browser (default is ``True``)
        return_fig : bool, optional
            Whether to return the figure (default is ``False``)

        Returns
        -------
        go.Figure
            An animated Plotly figure depicting the polygon file.  Returned if
            and only if ``return_fig`` is ``True``
        """
        # Verify that all polygons have the same units and determine range of
        # coordinates
        units = None
        xmin = math.inf
        xmax = -math.inf
        ymin = math.inf
        ymax = -math.inf

        for t, layer in self._polygon_data.items():
            for polygon in layer:
                if polygon.units is None:
                    raise PolygonFileMissingDataError(
                        'Units were not provided for polygon at time '
                        f'{t} {self.time_units}')

                if units is None:
                    units = polygon.units
                elif polygon.units != units:
                    raise PolygonFileFormatError(
                        'Polygons must have the same units to be plotted. '
                        f'Polygon at time {t} {self.time_units} has units '
                        f'{polygon.units} but previous polygons had units '
                        f'{units}')

                x, y = polygon.xy_coordinates()
                xmin = min(xmin, x.min())
                xmax = max(xmax, x.max())
                ymin = min(ymin, y.min())
                ymax = max(ymax, y.max())

        x_pad = 0.05*(xmax - xmin)
        y_pad = 0.05*(ymax - ymin)

        # Determine maximum number of traces per layer (since Plotly
        # animations require all frames to have the same number of traces)
        max_layer_traces = 0
        for layer in self._polygon_data.values():
            layer_traces = 0

            for polygon in layer:
                if polygon.construction:
                    layer_traces += 1
                else:
                    layer_traces += 2

            max_layer_traces = max(max_layer_traces, layer_traces)

        # Create frames
        first_layer_fig = _create_blank_plotly_figure()
        frames = []
        for i, (t, layer) in enumerate(self._polygon_data.items()):
            layer_fig: go.Figure = layer.plot(units=units, show=False,
                                              return_fig=True)

            for _ in range(max_layer_traces - len(layer_fig.data)):
                layer_fig.add_trace(go.Scatter(x=[], y=[]))

            frames.append(
                go.Frame(data=layer_fig.data, layout=layer_fig.layout, name=t))

            if i == 0:
                first_layer_fig = layer_fig

        # Create main figure
        figure = go.Figure(data=first_layer_fig.data,
                           layout=first_layer_fig.layout,
                           frames=frames)

        frame_args = {
            'frame': {'duration': delay},
            'mode': 'immediate',
            'fromcurrent': True,
            'transition': {'duration': 0},
        }

        sliders = [{
            'steps': [
                {
                    'args': [[frame.name], frame_args],
                    'label': f'{float(frame.name):.8g}',
                    'method': 'animate',
                } for frame in frames
            ],

            # Styling
            'x': 0.05,
            'y': 0,
            'len': 0.95,
            'pad': {'b': 10, 't': 60},
        }]

        figure.update_layout(
            xaxis={'range': [xmin-x_pad, xmax+x_pad], 'autorange': False},
            yaxis={'range': [ymin-y_pad, ymax+y_pad], 'autorange': False},
            updatemenus=[{
                'buttons': [
                    {
                        'args': [None, frame_args],
                        'label': '&#9654;',
                        'method': 'animate',
                    },
                    {
                        'args': [[None], frame_args],
                        'label': '<b>||</b>',
                        'method': 'animate',
                    },
                ],
                'direction': 'left',
                'type': 'buttons',

                # Styling
                'x': 0.03,
                'y': 0,
                'pad': {'r': 10, 't': 80},
            }],
            sliders=sliders,
        )

        if show:
            figure.show(config=_figure_config)  # pragma: no cover

        if return_fig:
            return figure

        return None

    def time_step(self, units: Optional[str] = None) -> float:
        """Returns the time step for the polygon file

        Verifies that the polygon file has a constant, positive time step between
        successive values in :py:attr:`time_values` and returns this time step.
        If the polygon file has only a single time step, ``0`` is returned.

        Parameters
        ----------
        units : str, optional
            The units in which the time step should be returned.  Required to
            be provided if the polygon file has more than one time step, but
            optional otherwise (default is ``None``)

        Returns
        -------
        float
            The time step between successive values in :py:attr:`time_values`,
            or ``0`` for polygon files with a single time step

        Raises
        ------
        PolygonFileFormatError
            If the polygon file has multiple time values but the time step
            between them is not consistent

        Notes
        -----
        The term "time" is used loosely for polygon files, and "time" may also
        be given in terms of quantities such as shaft rotation angle.  For
        more information, refer to the :ref:`fileref-polygon_file` page.
        """
        if self.num_time_steps <= 1:
            return 0

        if units is None:
            raise TypeError(
                f'Polygon file has {self.num_time_steps} time steps, so '
                'argument "units" must be provided')

        fp_tolerance = 1e-12

        time_vals = np.array(self.time_values, dtype=np.float64)
        time_steps = time_vals[1:] - time_vals[:-1]

        mean_time_step = float(np.mean(time_steps))
        max_diff = float(np.max(np.abs(time_steps - mean_time_step)))

        if max_diff > fp_tolerance:
            raise PolygonFileFormatError(
                'Inconsistent time step in polygon file. The mean time step is '
                f'{mean_time_step} {self.time_units} but individual time steps '
                f'differ by up to {max_diff} {self.time_units}'
            )

        time_step = float(self.unit_converter.convert(
            quantity=mean_time_step,
            from_unit=self.time_units, to_unit=units))

        if time_step <= 0:
            raise PolygonFileFormatError(
                'Polygon file time step cannot be negative or zero '
                f'(calculated time step is {time_step} {self.time_units})')

        return time_step

    def update_contents(self) -> None:
        """Updates the :py:attr:`contents` list based on object attributes

        This method synchronizes the :py:attr:`contents` list with the data
        stored in the :py:class:`PolygonFile` attributes.  This step should
        generally be performed prior to calling :py:meth:`write`, as writing
        a file directly saves the data in :py:attr:`contents` to disk.

        Construction polygons are not added to the :py:attr:`contents` list.
        """
        self.contents.clear()

        # Determine number of non-construction polygons per time step
        num_polygons = np.array(
            [sum(not shape.construction for shape in x)
             for x in self._polygon_data.values()])
        polygons_per_time_step = num_polygons[0]

        if not np.all(num_polygons == polygons_per_time_step):
            raise PolygonFileFormatError(
                'Different numbers of polygons are present at different time '
                'steps. To write a polygon file, all time steps must have the '
                'same number of polygons')

        # Write file
        self.contents.append(f'{self.num_time_steps} {polygons_per_time_step} '
                             f'{self.polygon_merge_method}')

        if self.num_time_steps > 1:
            time_step = self.time_step(self.time_units)

            self.contents.append(f'{self.time_units}: {self.time_values[0]} '
                                 f'{time_step} {self.time_extrap_method}')

        for t, layer in self._polygon_data.items():
            for polygon in layer:
                if polygon.construction:
                    continue

                if polygon.units is None:
                    raise PolygonFileMissingDataError(
                        'Units were not provided for polygon at time '
                        f'{t} {self.time_units}')

                if not isinstance(polygon, ClosedShape2D):
                    raise PolygonFileFormatError(
                        f'Polygon at time {t} {self.time_units} is not a '
                        f'subclass of `{ClosedShape2D.__name__}`')

                self.contents.append(f'{len(polygon.points())} '
                                     f'{polygon.polygon_file_enclosed_conv}')

                for coordinates in polygon.xy_coordinates(repeat_end=False):
                    coordinates_str = [str(x) for x in coordinates]
                    self.contents.append(
                        f'{polygon.units}: {" ".join(coordinates_str)}')



"""Utilities for loading simulation results and SimViewer configuration files.
"""

import base64
import json
from typing import Tuple

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash  # type: ignore
from packaging.version import Version

from mahautils.multics.simresults import SimResults
from .constants import GUI_SHORT_NAME, PROJECT_NAME, VERSION


def decode_base64(base64_str: str) -> str:
    """Decodes Base64 binary data with UTF-8 encoding"""
    content_data = base64_str.split(',', maxsplit=1)[1]
    return base64.b64decode(content_data).decode('utf_8')


## SIMULATION RESULTS FILES ##

def load_simresults(dash_base64_contents: str) -> SimResults:
    """Loads a simulation results file uploaded by Dash a :py:class:`dcc.Upload`
    object"""
    sim_results = SimResults()
    sim_results.set_contents(decode_base64(dash_base64_contents).split('\n'),
                             trailing_newline=True)
    sim_results.parse()

    return sim_results


## PLOT CONFIGURATION FILES ##

def plot_config_to_str(config_general: dict, config_x: dict, config_y: dict):
    """Combines plot configuration settings (general, x-axis, y-axes) into a
    single dictionary and exports it as a string"""
    combined_data = {
        'version': VERSION,
        'general': config_general,
        'x': config_x,
        'y': config_y,
    }

    return json.dumps(combined_data, indent=4)


def load_plot_config(dash_base64_contents: str) -> Tuple[dict, dict, dict]:
    """Loads plot configuration settings (general, x-axis, y-axes) from a file
    uploaded by Dash a :py:class:`dcc.Upload` object"""
    try:
        combined_data = json.loads(decode_base64(dash_base64_contents))
    except json.JSONDecodeError as exception:
        exception.args = ('Plot configuration file is not a valid JSON file',)
        raise

    try:
        version = combined_data['version']
        config_general = combined_data['general']
        config_x = combined_data['x']
        config_y = combined_data['y']
    except KeyError as exception:
        exception.args = (
            'Invalid plot configuration JSON file. The file does not contain '
            f'required key "{exception.args[0]}"',)
        raise

    minimum_version = '0.1.0'
    if Version(version) < Version(minimum_version):
        raise ValueError(
            'The plot configuration file you uploaded was generated with '
            f'{PROJECT_NAME} v{version}, but the minimum required version '
            f'that can be read is v{minimum_version}')

    maximum_version = VERSION
    if Version(version) > Version(maximum_version):
        raise ValueError(
            'The plot configuration file you uploaded was generated with '
            f'{PROJECT_NAME} v{version}, but the maximum permitted version '
            f'that can be read is v{maximum_version}')

    return config_general, config_x, config_y


load_plot_config_error_message = dash.html.Div([
    dash.html.P(
        'The plot configuration file you uploaded is not valid.  This typically '
        'occurs either because the file does not have the required format or '
        'the simulation results files it specifies have not been uploaded.'
    ),
    dash.html.Br(),
    dash.html.P(
        'To avoid issues, is generally recommended that you generate the plot '
        f'configuration file using the "Export" button in the {GUI_SHORT_NAME} '
        'GUI and do not manually edit it to avoid formatting issues.  Also, it '
        'is advisable to upload your simulation results files BEFORE importing '
        'a plot configuration file.'
    ),
])



"""This module provides a base class intended to aid in reading, parsing, and
writing Maha Multics configuration files.
"""

import pathlib
import re
from typing import List, Optional, Tuple, Union

import pyxx

from .exceptions import (
    FileNotParsedError,
    MahaMulticsFileFormatError,
)
from .units import MahaMulticsUnitConverter


class MahaMulticsConfigFile(pyxx.files.TextFile):
    """A generic class for processing Maha Multics configuration files

    This class is intended to represent a range of Maha Multics configuration
    files, and configures settings (such as the character used for comments)
    that are applicable to all Maha Multics configuration files, as well as
    providing general methods for processing such files.  Subclasses should
    generally be created and customized to specific types of Maha Multics
    files.
    """

    def __init__(self, path: Optional[Union[str, pathlib.Path]] = None,
                 unit_converter: Optional[pyxx.units.UnitConverter] = None
                 ) -> None:
        """Creates an object to represent Maha Multics configuration files

        Creates an instance of the :py:class:`MahaMulticsConfigFile` class,
        including configuring file comments to be represented by the ``#``
        character.

        Parameters
        ----------
        path : str or pathlib.Path, optional
            Location of the text file in the file system  (default is ``None``)
        unit_converter : pyxx.units.UnitConverter, optional
            A :py:class:`pyxx.units.UnitConverter` instance which will be
            used to convert units of quantities stored in the configuration file
            (default is ``None``, which uses the
            :py:class:`MahaMulticsUnitConverter` unit converter to perform unit
            conversions)
        """
        super().__init__(path=path, comment_chars='#')

        self.unit_converter = MahaMulticsUnitConverter() \
            if unit_converter is None else unit_converter

    @property
    def unit_converter(self) -> pyxx.units.UnitConverter:
        """The unit converter used to convert the units of quantities stored
        in the file"""
        return self._unit_converter

    @unit_converter.setter
    def unit_converter(self, unit_converter: pyxx.units.UnitConverter) -> None:
        if not isinstance(unit_converter, pyxx.units.UnitConverter):
            raise TypeError('Argument "unit_converter" must be of type '
                            f'{pyxx.units.UnitConverter}')

        self._unit_converter = unit_converter

    def _extract_full_line_comment_text(self, line: str) -> str:
        """Strips whitespace and leading comment characters from a string

        Parameters
        ----------
        line : str
            The line of the file to process

        Returns
        -------
        str
            The line with any of :py:attr:`comment_chars` removed from the
            beginning of the line and leading and trailing whitespace stripped
        """
        line = line.strip()

        # Since `comment_chars` is hard-coded to be `('#',)` for this class,
        # this condition should always be true; however, this check was still
        # added to be safe
        if self.comment_chars is not None:  # pragma: no cover
            while line.startswith(self.comment_chars):
                for char in self.comment_chars:
                    line = line.lstrip(char).strip()

        return line.strip()

    def extract_section_by_keyword(self, section_label: str,
                                   begin_regex: str, end_regex: str,
                                   section_line_regex: str = r'(.*)',
                                   max_sections: Optional[int] = None,
                                   begin_idx: int = 0,
                                   allow_comment_lines: bool = True,
                                   ) -> Tuple[List[re.Match],
                                              List[Tuple[str, ...]],
                                              int,
                                              int]:
        """Extracts a section from the :py:attr:`contents` list of file lines

        Many Maha Multics configuration files contain sections with certain
        types of data, where the section begins following a formatted section
        marker and ends at another marker (both with unique, identifiable
        regex patterns).  This method extracts the data from such a section.
        If multiple sections are found, the data in all sections is merged,
        unless specified otherwise by setting ``max_sections``.

        Parameters
        ----------
        section_label : str
            A descriptive name identifying the section.  This is not used in
            parsing the file; it is only used to customize error messages and
            make them more descriptive
        begin_regex : str
            The regex pattern which marks the beginning of the section
        end_regex : str
            The regex pattern which marks the end of the section
        section_line_regex : str, optional
            If provided, this regex pattern must be matched by all lines
            inside the section (default is ``'(.*)'``, which matches any text)
        max_sections : int, optional
            The maximum number of sections to extract; that is, only the first
            ``max_sections`` encountered will be extracted and returned
            (default is ``None``, which extracts all sections)
        begin_idx : int, optional
            The index (in the :py:attr:`contents` list) at which to begin to
            search for and extract data from sections (default is ``0``)
        allow_comment_lines : bool, optional
            If ``True``, any lines within the section that do not match
            ``section_line_regex`` but begin with any of the characters in
            :py:attr:`comment_chars` will be outputted (part of the second
            output of the method); if ``False``, any lines within the section
            that do not match ``section_line_regex`` will result in an error
            being thrown (default is ``True``)

        Returns
        -------
        list
            A list of :py:class:`re.Match` objects containing the matches for
            the regex pattern ``section_line_regex`` for all lines in the
            section(s)
        list
            A list (of the same length as the first argument returned) of
            tuples of strings.  For each :py:class:`re.Match` object, the
            corresponding item in this list contains a tuple with any full-line
            comments preceding the matched line
        int
            The index of :py:attr:`contents` of the next line immediately
            following the line on which ``end_regex`` was found
        int
            The number of sections that were extracted from the
            :py:attr:`contents` list
        """
        # Validate inputs
        if (max_sections is not None) \
                and (max_sections := int(max_sections)) <= 0:
            raise ValueError('Argument "max_sections" must be a positive integer')

        # Counter to track how many sections have been found
        num_sections = 0

        # Lists to store all regex groups in the section and associated comments
        section_regex_groups = []
        section_line_comments = []

        # Iterate through file and extract section contents
        i = begin_idx
        while i < len(self.contents):
            comment_lines: List[str] = []

            if (max_sections is not None) and (num_sections >= max_sections):
                break

            line = self.contents[i]

            # Identify lines beginning with a given regex pattern, and
            # extract all lines matching a particular regex pattern until
            # another specified regex pattern is encountered
            if (groups := re.search(begin_regex + r'(.*)', line)) is not None:
                num_sections += 1

                # Check to see whether there are data to be extracted on the
                # same line as the beginning section identifier (and if not,
                # advance to next line)
                if (line := groups.groups()[-1]) == '':
                    i += 1
                    line = self.contents[i]

                while not re.match(end_regex, line):
                    groups = re.search(section_line_regex, line)

                    if groups is not None:
                        section_regex_groups.append(groups)
                        section_line_comments.append(tuple(comment_lines))
                        comment_lines.clear()

                    elif (
                        allow_comment_lines
                        and self.comment_chars is not None
                        and line.strip().startswith(self.comment_chars)
                    ):
                        comment_lines.append(line)

                    else:
                        raise MahaMulticsFileFormatError(
                            f'Line "{line}" in section "{section_label}" does '
                            'not match expected pattern')

                    i += 1

                    if i >= len(self.contents):
                        raise MahaMulticsFileFormatError(
                            f'Unable to locate end of "{section_label}" section')

                    line = self.contents[i]

            i += 1

        return (section_regex_groups, section_line_comments, i, num_sections)

    def parse(self) -> None:
        """Parses the data in :py:attr:`contents` and stores it in class
        attributes

        For Maha Multics configuration files, the :py:meth:`parse` method
        verifies that the file has been read prior to attempting to parse it.
        """
        super().parse()

        # Verify that file contents have been read
        if len(self.contents) == 0:
            raise FileNotParsedError(
                'Unable to parse file. File has not yet been read')



"""**Shapes that can be used when constructing geometry with MahaUtils**

This module provides geometric shapes that can be used when creating more
complex geometry (for instance, to write Maha Multics polygon files).  The
specification of the shapes in this module is intended to be as generic and
flexible as possible to allow greater freedom in building complex geometry.
"""

from .circle import Circle
from .geometry import Geometry
from .point import Point
from .point2D import CartesianPoint2D
from .point3D import CartesianPoint3D
from .polygon import Polygon
from .shape import Shape2D
from .shape_open_closed import ClosedShape2D, OpenShape2D



"""This module provides a class intended for reading and parsing VTK files
used by the Maha Multics software.
"""

import enum
import io
import pathlib
import re
import string
from typing import Any, Dict, List, Tuple, Union, Optional

# Mypy type checking disabled for packages that are not PEP 561-compliant
import numpy as np
import pandas as pd            # type: ignore
import pyxx
import scipy.interpolate       # type: ignore
import vtk                     # type: ignore
import vtk.util.numpy_support  # type: ignore  # pylint: disable=E0401,E0611

from mahautils.utils.capture_printing import CaptureStderr
from .exceptions import (
    FileNotParsedError,
    VTKIdentifierNameError,
    VTKInvalidIdentifierError,
    VTKFormatError,
)
from .units import MahaMulticsUnitConverter

# Type definitions
PointCoordinates = Union[List[float], Tuple[float, ...], np.ndarray]
ListOfPoints = Union[
    List[PointCoordinates],
    Tuple[PointCoordinates, ...],
    np.ndarray
]


class VTKDataType(enum.Enum):
    """The type of point data stored in a VTK file

    VTK files can store a variety of types of data (scalars, vectors, tensors,
    etc.).  This :py:class:`enum.Enum` specifies the different types that can
    be read by the MahaUtils package.
    """

    #: Scalar data series
    scalar = enum.auto()

    #: Vector data series
    vector = enum.auto()


class VTKFile(pyxx.files.BinaryFile):
    """An object for representing VTK files

    VTK files are commonly used with the Maha Multics software to store film
    property distributions and other simulation results.  This class provides
    the ability to read and parse such files, and perform tasks such as
    extracting particular data, converting the units of data stored in the VTK
    file, and plotting scalar property distributions.
    """

    # Regular expression describing expected format for VTK data identifiers
    # to facilitate unit conversions
    __unit_conversion_regex = r'^([^\s]+)\[([^\s]+)\]$'

    def __init__(self, path: Optional[Union[str, pathlib.Path]] = None,
                 unit_converter: Optional[pyxx.units.UnitConverter] = None,
                 **kwargs) -> None:
        """Creates an object that can parse data from a VTK file

        Creates an instance of the :py:class:`VTKFile` class and optionally
        reads and parses a specified VTK file.

        Parameters
        ----------
        path : str or pathlib.Path, optional
            The path and filename of the VTK file to read and parse (default
            is ``None``).  If set to ``None``, no VTK file is read
        unit_converter : pyxx.units.UnitConverter, optional
            A :py:class:`pyxx.units.UnitConverter` instance which will be
            used to convert units of quantities stored in the VTK file
            (default is ``None``).  If set to ``None``, the
            :py:class:`MahaMulticsUnitConverter` unit converter will be used
            to perform unit conversions
        **kwargs
            Any valid arguments (other than ``path``) for the :py:meth:`read`
            method can be passed to this constructor as keyword arguments
        """
        super().__init__(path)

        # Initialize variables
        self._coordinate_units: Union[str, None] = None
        self._vtk_data_types: Dict[str, VTKDataType] = {}

        if unit_converter is None:
            self.unit_converter = MahaMulticsUnitConverter()
        else:
            self.unit_converter = unit_converter

        # If file is specified, read the file
        if path is not None:
            self.read(path=path, **kwargs)

    @property
    def coordinate_units(self) -> Union[str, None]:
        """The units in which the :math:`x`-, :math:`y`-, and
        :math:`z`-coordinates of points in the VTK file are stored
        """
        return self._coordinate_units

    @property
    def identifiers(self) -> List[str]:
        """The list of all VTK data identifiers for data stored in the file"""
        return list(self._df.columns)

    @property
    def num_points(self) -> int:
        """The number of points in the VTK grid for which vector, scalar,
        and/or tensor data are stored
        """
        try:
            return self._num_points
        except AttributeError as exception:
            raise FileNotParsedError(
                'Number of points in VTK grid is not defined; VTK file has '
                'not yet been read') from exception

    @property
    def pointdata_df(self) -> pd.DataFrame:
        """A Pandas DataFrame containing the coordinates of each point in the
        VTK file and any scalar or vector data for the point

        This DataFrame stores the raw data parsed from the VTK file.  The
        first three columns of the file store the :math:`x`-, :math:`y`-, and
        :math:`z`-coordinates of point in the VTK grid, and subsequent columns
        store the raw scalar or vector data for each point.  Data are stored
        in the same units defined in the VTK file.
        """
        try:
            return self._df.copy(deep=True)
        except AttributeError as exception:
            raise FileNotParsedError(
                'Point data DataFrame is not defined; VTK file has not yet '
                'been read') from exception

    @property
    def unit_conversion_enabled(self) -> bool:
        """Whether the :py:class:`VTKFile` instance is capable of performing
        unit conversions on VTK data

        VTK files don't inherently store data.  However, it can be useful to
        perform unit conversions and extract data from VTK files in different
        units than the data were stored.  The :py:class:`VTKFile` provides
        such unit conversion capability, but in order to do so, the user must
        appropriately name the data identifiers in the VTK file such that they
        include the unit in which the data are stored.

        The naming convention adopted in this package to facilitate unit
        conversions for VTK data requires that VTK data identifiers (for both
        scalar and vector data) are formatted in two parts: (1) a descriptive
        name, (2) the unit in square brackets.  There should be no whitespace
        in any part of the identifier.

        For instance, one potential identifier that could denote VTK data
        storing pressure in units of Pascal might be: ``pressure[Pa]``.
        Similarly, an identifier for VTK data storing the velocity of a
        tractor might be ``tractorVelocity[m/s]``.
        """
        try:
            return self._unit_conversion_enabled
        except AttributeError as exception:
            raise FileNotParsedError(
                'VTK identifier naming convenction is not defined; VTK file '
                'has not yet been read') from exception

    @property
    def unit_converter(self) -> pyxx.units.UnitConverter:
        """The unit converter used to perform unit conversions for quantities
        stored in the VTK file

        This attribute must be an instance or subclass of a
        :py:class:`pyxx.units.UnitConverter` object.  Unit conversions are
        only performed if :py:attr:`unit_conversion_enabled` is ``True``.
        """
        return self._unit_converter

    @unit_converter.setter
    def unit_converter(self,
                       unit_converter: Optional[pyxx.units.UnitConverter]
                       ) -> None:
        if not isinstance(unit_converter, pyxx.units.UnitConverter):
            raise TypeError(
                'Argument "unit_converter" must be an instance or subclass '
                'of "pyxx.units.UnitConverter"')

        self._unit_converter = unit_converter

    def _check_unit_conversion_compliance_args(self, unit: Any) -> None:
        """Verifies that, when retrieving VTK data, the user specified or
        omitted a unit in agreement with the unit conversion settings
        specified by :py:attr:`unit_conversion_enabled`

        Several methods in the :py:class:`VTKFile` class retrieve VTK data
        and perform unit conversions if :py:attr:`unit_conversion_enabled`
        is ``True``.  These methods need to verify that a unit was provided
        when unit conversions are enabled, and to avoid user confusion and
        potential bugs (i.e., a user thinks a unit conversion is being
        performed but it is not), the method should verify that a unit was not
        provided if unit conversions are disabled.  Rather than repeating the
        code to perform this check throughout the :py:class:`VTKFile` class
        source code, this method can be called and will throw a descriptive
        error if any of the aforementioned conditions are not satisfied.

        Parameters
        ----------
        unit : Any
            The argument passed by the user specifying the unit to which VTK
            data should be converted (or ``None`` if no unit was specified)

        Raises
        ------
        TypeError
            If unit conversions are disabled but the user specified a unit, or
            if unit conversions are enabled but the user omitted a unit
        """
        if self.unit_conversion_enabled:
            if unit is None:
                raise TypeError(
                    'VTK unit conversions are enabled, so argument "unit" '
                    'cannot be `None`')

        elif unit is not None:
            raise TypeError(
                'VTK unit conversions are not enabled, so argument "unit" '
                'must be `None`')

    def _check_unit_conversion_compliance_id(self, identifier: str) -> None:
        """Verifies that a given VTK data identifier matches the naming
        convention required by :py:attr:`unit_conversion_enabled`

        VTK data identifiers must be strings, and if unit conversions are
        enabled (:py:attr:`unit_conversion_enabled` is ``True``), then the
        VTK data identifiers must follow a specific format including the unit.
        This method confirms that these requirements are met, and throws a
        descriptive error if they are not.

        Parameters
        ----------
        identifier : str
            The VTK data identifier (i.e., the column name of the VTK data
            DataFrame stored in :py:attr:`pointdata_df`) to be checked

        Raises
        ------
        TypeError
            If ``identifier`` is not of type ``str``
        VTKIdentifierNameError
            If :py:attr:`unit_conversion_enabled` is ``True`` and
            ``identifier`` does not match the required format (described for
            the :py:attr:`unit_conversion_enabled` attribute)
        """
        if not isinstance(identifier, str):
            raise TypeError(
                f'VTK data name {identifier} is not of type "str"')

        if self.unit_conversion_enabled:
            if not re.match(self.__unit_conversion_regex, identifier):
                raise VTKIdentifierNameError(
                    f'Invalid VTK data identifier: "{identifier}"')

    def _find_column_id(self, identifier: str) -> str:
        # Validate inputs
        if not isinstance(identifier, str):
            raise TypeError('Argument "identifier" must be of type "str"')

        # If unit conversions are disabled, simply check that the
        # identifier exists as one of the DataFrame columns
        if not self.unit_conversion_enabled:
            if identifier not in self._df:
                raise VTKInvalidIdentifierError(
                    f'Data specified by identifier "{identifier}" not found '
                    'in VTK file')

        # If unit conversions are enabled, users can specify the identifier
        # with or without the unit (e.g., 'pFilm[bar]' or 'pFilm'), so
        # determine which of these is a valid column name (and if neither,
        # throw an error)
        else:
            if identifier not in self._df:
                matches = []
                for column in self._df.columns:
                    if identifier == self._parse_column_id(column, 'name'):
                        matches.append(column)

                if len(matches) == 0:
                    raise VTKInvalidIdentifierError(
                        f'Data specified by identifier "{identifier}" not found '
                        'in VTK file')

                if len(matches) > 1:
                    # Ideally, it should never be possible to reach this
                    # statement because an error would have been thrown while
                    # reading the VTK file.  However, this check is still
                    # implemented as a backup precaution

                    raise VTKInvalidIdentifierError(  # pragma: no cover
                        f'Identifier "{identifier}" matches multiple data '
                        f'fields in the VTK file: {matches}')

                # At this point, there must be only a single item in `matches`,
                # corresponding to the column name of the VTK data identifier
                identifier = matches[0]

        return identifier

    def _parse_column_id(self, identifier: str, target: str) -> str:
        if not self.unit_conversion_enabled:
            raise AttributeError(
                'Parsing VTK data identifier is not a valid action when VTK '
                'unit conversions are disabled')

        matches = re.search(self.__unit_conversion_regex, str(identifier))

        if matches is None:
            raise VTKIdentifierNameError(
                f'Invalid VTK data identifier: "{identifier}"')

        if target == 'name':
            return matches.group(1)
        if target == 'unit':
            return matches.group(2)

        raise ValueError('Argument "target" must be one of: ["name", "unit"]')

    def coordinates(self, axis: str, unit: Optional[str] = None) -> np.ndarray:
        """Returns a NumPy array containing the coordinates of all grid points
        in the VTK file along a particular coordinate axis

        VTK files store data (scalars, vectors, etc.) at a set of defined grid
        points in 3D.  This method returns a 1D NumPy array containing the
        coordinates of the coordinates of such points along an axis specified
        by ``axis``.  Point coordinates are returned in the order in which
        they were defined in the VTK file.

        Parameters
        ----------
        axis : str
            The coordinate axis for which to retrieve coordinates.  Must be
            exactly one of: ``'x'``, ``'y'``, ``'z'``
        unit : str, optional
            The unit in which the VTK points should be returned (default is
            ``None``).  Must be provided if :py:attr:`unit_conversion_enabled`
            is ``True`` and omitted if :py:attr:`unit_conversion_enabled` is
            ``False``

        Returns
        -------
        np.ndarray
            A 1D NumPy array containing the coordinates of all VTK grid points
            along the axis specified by ``axis``

        Raises
        ------
        FileNotParsedError
            If attempting to call this method before calling :py:meth:`read`
            to read and parse a VTK file
        """
        try:
            if axis.lower() == 'x':
                column = self.__xyz_coordinate_columns[0]
            elif axis.lower() == 'y':
                column = self.__xyz_coordinate_columns[1]
            elif axis.lower() == 'z':
                column = self.__xyz_coordinate_columns[2]
            else:
                raise ValueError(
                    'Argument "axis" must be exactly one of: ["x", "y", "z"]')
        except AttributeError as exception:
            raise FileNotParsedError(
                'Cannot retrieve VTK x-coordinates; VTK file has not yet '
                'been read') from exception

        return self.extract_data_series(column, unit)

    def extract_data_series(self, identifier: str, unit: Optional[str] = None
                            ) -> np.ndarray:
        """Returns a NumPy array containing a single VTK data field

        VTK files store data (scalars, vectors, etc.) at a set of defined grid
        points in 3D.  This method retrieves a single such field of data (i.e.,
        it retrieves one column in :py:attr:`pointdata_df`), and returns the
        resulting values in a NumPy array.

        Parameters
        ----------
        identifier : str
            The identifier specifying the data in the VTK file to return
        unit : str, optional
            The units in which the data should be returned (only applicable if
            :py:attr:`unit_conversion_enabled` is ``True``; otherwise, must
            not be specified)

        Returns
        -------
        np.ndarray
            A NumPy array containing the data corresponding to ``identifier``
            in the VTK file
        """
        # SETUP --------------------------------------------------------------
        # Verify that file has been read
        if not hasattr(self, '_df'):
            raise FileNotParsedError(
                'Unable to retrieve VTK data; VTK file has not yet '
                'been read')

        # Check that identifier matches a single column in the DataFrame
        identifier = self._find_column_id(identifier)

        # Verify that unit was provided if and only if unit conversions
        # are enabled
        self._check_unit_conversion_compliance_args(unit)

        # Extract raw data from DataFrame
        raw_data = self._df[identifier].to_numpy()

        if raw_data.dtype != np.float64:
            # For vector data, the Pandas `.to_numpy()` method returns a
            # NumPy array with `dtype=object`.  Vector data should be an array
            # of floating-point numbers, which is performed here
            raw_data = np.array([x.astype(np.float64) for x in raw_data])

        # CASE 1: Unit conversions disabled ----------------------------------
        if not self.unit_conversion_enabled:
            return raw_data

        # CASE 2: Unit conversions enabled -----------------------------------
        # Convert raw data units
        from_unit = self._parse_column_id(identifier, 'unit')
        return self.unit_converter.convert(
            raw_data, from_unit=from_unit, to_unit=str(unit))

    def extract_dataframe(self, identifiers: List[str],
                          units: Optional[List[str]] = None) -> pd.DataFrame:
        """Returns a Pandas DataFrame containing one or more VTK data fields

        VTK files store data (scalars, vectors, etc.) at a set of defined grid
        points in 3D.  This method retrieves one or more such fields of data
        (i.e., it retrieves one or more columns in :py:attr:`pointdata_df`),
        and returns the resulting values in a Pandas DataFrame.

        Parameters
        ----------
        identifiers : list of str
            The (one or more) identifiers specifying the data in the VTK file
            to return
        units : list of str, optional
            The units in which the data should be returned (only applicable if
            :py:attr:`unit_conversion_enabled` is ``True``; otherwise, must
            not be specified).  If supplied, ``units`` should be a list of
            strings of equal length as ``identifiers``

        Returns
        -------
        pd.DataFrame
            A Pandas DataFrame containing the columns of :py:attr:`pointdata_df`
            corresponding to ``identifiers`` in the VTK file
        """
        # SETUP --------------------------------------------------------------
        # Verify that file has been read
        if not hasattr(self, '_df'):
            raise FileNotParsedError(
                'Unable to retrieve VTK data; VTK file has not yet '
                'been read')

        # Verify that unit was provided if and only if unit conversions
        # are enabled
        self._check_unit_conversion_compliance_args(units)

        # CASE 1: Unit conversions disabled ----------------------------------
        if not self.unit_conversion_enabled:
            df_data = {
                identifier: list(self.extract_data_series(identifier))
                    for identifier in identifiers  # noqa: E131
            }

        # CASE 2: Unit conversions enabled -----------------------------------
        else:
            # Ensure that units are a list of strings
            #   Add Mypy exclusion because if "units" is `None`, an error
            #   would have been thrown by unit conversion compliance check
            units = [str(unit) for unit in units]  # type: ignore

            # Validate inputs
            if len(identifiers) != len(units):
                raise ValueError(
                    'Arguments "identifiers" and "units" must be lists of '
                    'strings with the same number of items')

            # Extract and convert units of data from VTK file
            df_data = {}
            for i, identifier in enumerate(identifiers):
                name = self._parse_column_id(self._find_column_id(identifier),
                                             target='name')
                unit = units[i]

                df_data[f'{name}[{unit}]'] \
                    = list(self.extract_data_series(identifier, unit))

        return pd.DataFrame(df_data)

    def interpolate(self, identifier: str,
                    query_points: ListOfPoints,
                    interpolator_type: str,
                    output_units: Optional[str] = None,
                    query_point_units: Optional[str] = None,
                    interpolate_axes:
                        Union[List[str], Tuple[str, ...], str]
                        = ('x', 'y', 'z'),
                    idx_slice: Union[slice, tuple] = slice(None),
                    **kwargs
                    ) -> np.ndarray:
        """Interpolates data from the VTK file

        Retrieves the value of a given data field stored in the VTK file,
        interpolating between VTK grid points if necessary.  Interpolation is
        performed using the :py:mod:`scipy.interpolate` package
        (https://docs.scipy.org/doc/scipy/reference/interpolate.html).

        Parameters
        ----------
        identifier : str
            The identifier specifying the data in the VTK file to return
        query_points : tuple or list or np.ndarray
            The point(s) at which to return possibly interpolated value(s) of
            the VTK data corresponding to ``identifier``
        interpolator_type : str
            The SciPy interpolation function to use to perform interpolation.
            Can be selected from any of the options in the "Notes" section
        output_units : str, optional
            The units in which the data should be returned (only applicable if
            :py:attr:`unit_conversion_enabled` is ``True``; otherwise, must
            not be specified)
        query_point_units : str, optional
            The units of the ``query_points`` argument (only applicable if
            :py:attr:`unit_conversion_enabled` is ``True``; otherwise, must
            not be specified)
        interpolate_axes : list or tuple or set or str, optional
            The coordinate axes on which interpolation should be performed.
            Must be selected from any combination of ``'x'``, ``'y'``, and
            ``'z'`` (default is ``('x', 'y', 'z')``)
        idx_slice : slice or tuple, optional
            Filters the points in the VTK file and only uses a subset of
            points for interpolation (default is ``slice(None)`` which uses
            all points in the VTK file).  See the "Notes" section for more
            information
        **kwargs
            Any keyword arguments to be supplied to the SciPy interpolation
            function specified by ``interpolator_type``.  See the "Notes"
            section for more information

        Returns
        -------
        np.ndarray
            The interpolated value(s) of the VTK data given by ``identifier``
            at the query points ``query_points``

        Notes
        -----
        **Interpolation Functions**

        The following interpolation functions are available (set the
        ``interpolator_type`` argument to the given string to use each):

        - ``'griddata'``: The :py:class:`scipy.interpolate.griddata`
          interpolation function for unstructured, multivariate interpolation
          (`reference <https://docs.scipy.org/doc/scipy/reference/generated
          /scipy.interpolate.griddata.html>`__)
        - ``'RBFInterpolator'``: The :py:class:`scipy.interpolate.RBFInterpolator`
          function for unstructured, multivariate interpolation using a radial
          basis function (`reference <https://docs.scipy.org/doc/scipy/reference
          /generated/scipy.interpolate.RBFInterpolator.html>`__)

        Review the SciPy documentation for questions about parameters for the
        interpolation functions.  **These interpolation functions require data
        and interpolation parameters to meet specific mathematical requirements,
        and errors may be encountered if such requirements are not met.**  This
        may require using non-default parameters of the :py:meth:`interpolate`
        method.

        For instance, if your VTK file is defined on a 2D ``xy``-plane and you
        attempt to perform 3D interpolation (i.e., you set ``interpolation_axes``
        to ``('x', 'y', 'z')``) using ``griddata`` with ``method='linear'``, an
        error will be thrown.  In this case, you need to reduce the problem to
        a 2D interpolation by setting ``interpolation_axes`` to ``('x', 'y')``.

        **Point Coordinates**

        The order in which ``interpolate_axes`` values are provided must be in
        the following sequence: ``(x, y, z)``.  An error will be thrown if an
        argument such as ``interpolate_axes=('y', 'x', 'z')`` is provided.

        **Filtering/Slicing Points**

        In some cases, it may be desirable to perform interpolation using only
        a subset of points in the VTK file.  For instance, if a lubricating
        film lies on a specific face, it may be desirable to only perform
        interpolation using that face.

        The ``idx_slice`` argument facilitates this use case.  Either a Python
        :py:class:`slice` object can be passed as input, or a tuple of array
        indices generated by NumPy's :py:meth:`index_exp` or :py:meth:`np.s_`
        methods (`more information <https://numpy.org/doc/stable/reference
        /generated/numpy.s_.html>`__).

        Note that if using :py:meth:`index_exp` or :py:meth:`np.s_`, only a
        1D index tuple should be generated (e.g., ``np.index_exp[0:4]``).  An
        easy way to test the index tuple is to apply it to the output of
        :py:meth:`points` (e.g., ``vtk_file.points()[np.index_exp[...]]``)
        and observe whether the desired points are extracted (these are the
        points that would be used for interpolation).
        """
        # Validate inputs
        axes = list(interpolate_axes)
        num_axes = len(axes)

        allowed_axes = {'x', 'y', 'z'}
        if not set(axes).issubset(allowed_axes):
            raise ValueError(
                'Argument "interpolate_axes" contains invalid items: '
                f'{set(axes) - allowed_axes}')

        if len(set(axes)) != len(axes):
            raise ValueError(
                'Argument "interpolate_axes" contains duplicate values')

        if not (1 <= num_axes <= 3):
            raise ValueError('Number of interpolation dimensions must be '
                             'between 1 and 3 (inclusive)')

        if not axes == sorted(axes):
            raise ValueError(
                'Argument "interpolate_axes" must be in the following '
                f'sequence: {sorted(axes)}')

        self._check_unit_conversion_compliance_args(output_units)
        self._check_unit_conversion_compliance_args(query_point_units)

        try:
            # Convert query points to NumPy array
            query_points_ndarray = np.array(query_points, dtype=np.float64)

            if query_points_ndarray.ndim == 1:
                query_points_ndarray = np.expand_dims(query_points_ndarray,
                                                      axis=0)

            # Check that query points have expected shape
            if not query_points_ndarray.ndim == 2:
                raise ValueError('Query points must be provided as a 2D array')

            if not query_points_ndarray.shape[1] == num_axes:
                raise ValueError(
                    f'Query points must each have {num_axes} coordinates '
                    f'({", ".join(sorted(axes))})')

        except ValueError as exception:
            raise ValueError(
                'Invalid format of "query_points" argument. This argument '
                f'should be provided as a shape (N,{num_axes}) NumPy array '
                'of N points') from exception

        if not isinstance(idx_slice, slice):
            if not isinstance(idx_slice, tuple):
                raise TypeError(
                    'Argument "idx_slice" must be of type "slice" or "tuple"')

            if len(idx_slice) != 1:
                raise ValueError('Argument "idx_slice" is of type "tuple" '
                                 'but length is not 1')

        # Extract point coordinates of all points in VTK file
        coordinates = []
        if 'x' in axes:
            coordinates.append(self.coordinates('x', query_point_units)[idx_slice])
        if 'y' in axes:
            coordinates.append(self.coordinates('y', query_point_units)[idx_slice])
        if 'z' in axes:
            coordinates.append(self.coordinates('z', query_point_units)[idx_slice])

        points = np.stack(coordinates, axis=1)

        # Extract scalar data for all points in VTK file
        scalar_data = self.extract_data_series(identifier, output_units)[idx_slice]

        # Perform interpolation
        if (interpolator_type == 'griddata'):
            return scipy.interpolate.griddata(
                points = points,
                values = scalar_data,
                xi     = query_points_ndarray,
                **kwargs
            )

        if (interpolator_type == 'RBFInterpolator'):
            interpolator = scipy.interpolate.RBFInterpolator(
                y = points,
                d = scalar_data,
                **kwargs
            )

            return interpolator(query_points_ndarray)

        raise ValueError(f'Interpolator "{interpolator_type}" is not supported')

    def is_scalar(self, identifier: str) -> bool:
        """Whether a given VTK data identifier stores scalar point data

        Parameters
        ----------
        identifier : str
            The VTK data identifier to analyze

        Returns
        -------
        bool
            Returns ``True`` if the VTK data identifier given by ``identifier``
            stores scalar data, and ``False`` otherwise
        """
        column = self._find_column_id(identifier)
        return self._vtk_data_types[column] == VTKDataType.scalar

    def is_vector(self, identifier: str) -> bool:
        """Whether a given VTK data identifier stores vector point data

        Parameters
        ----------
        identifier : str
            The VTK data identifier to analyze

        Returns
        -------
        bool
            Returns ``True`` if the VTK data identifier given by ``identifier``
            stores vector data, and ``False`` otherwise
        """
        column = self._find_column_id(identifier)
        return self._vtk_data_types[column] == VTKDataType.vector

    def points(self, unit: Optional[str] = None) -> np.ndarray:
        """Returns a list of all grid points in the VTK file

        VTK files store data (scalars, vectors, etc.) at a set of defined grid
        points in 3D.  This method returns a list containing the coordinates
        of all such points.  Refer to the "Notes" section for details about
        the format of the returned points.

        Parameters
        ----------
        unit : str, optional
            The unit in which the VTK points should be returned (default is
            ``None``).  Must be provided if :py:attr:`unit_conversion_enabled`
            is ``True`` and omitted if :py:attr:`unit_conversion_enabled` is
            ``False``

        Returns
        -------
        np.ndarray
            A NumPy array containing a list of all VTK grid points.  Refer to
            the "Notes" section for details about the format of the array

        Raises
        ------
        FileNotParsedError
            If attempting to call this method before calling :py:meth:`read`
            to read and parse a VTK file

        Notes
        -----
        The VTK grid points are returned as a 2D array, where the first index
        specifies a particular point (out of the :py:attr:`num_points` points)
        and the second index specifies the coordinate axis (:math:`x`,
        :math:`y`, or :math:`z`).

        For example, suppose that the VTK file stored data for five points:
        ``(x1, y1, z1)``, ``(x2, y2, z2)``, ..., ``(x5, y5, z5)``.  In this
        case, the :py:meth:`points` method would return:

        .. code-block:: python

            array([[x1, y1, z1],
                   [x2, y2, z2],
                   [x3, y3, z3],
                   [x4, y4, z4],
                   [x5, y5, z5]])

        Point coordinates are returned in the order in which they were defined
        in the VTK file.
        """
        try:
            return np.array([self.extract_data_series(name, unit)
                             for name in self.__xyz_coordinate_columns]
                            ).transpose()
        except AttributeError as exception:
            raise FileNotParsedError(
                'Cannot retrieve VTK grid points; VTK file has not yet '
                'been read') from exception

    def read(self,
             path: Optional[Union[str, pathlib.Path]] = None,
             unit_conversion_enabled: bool = False,
             coordinate_units: Optional[str] = None,
             strict: bool = False,
             fallback_units: Optional[Dict[str, str]] = None
             ) -> None:
        """Reads a VTK file from the disk

        This method reads a VTK file from the disk, parsing its content and
        storing the data as a Pandas DataFrame in the :py:attr:`pointdata_df`
        attribute.

        Parameters
        ----------
        path : str or pathlib.Path, optional
            The path and filename from which to read the VTK file (default is
            ``None``).  If not provided or ``None``, the file will be read
            from the location specified by the :py:attr:`path` attribute
        unit_conversion_enabled : bool, optional
            Whether to enable unit conversions for the VTK file data (see
            the :py:attr:`unit_conversion_enabled` attribute for additional
            details) (default is ``False``)
        coordinate_units : str, optional
            The units used by the coordinate system in the VTK file (default
            is ``None``).  Must be provided if :py:attr:`unit_conversion_enabled`
            is ``True`` and omitted if :py:attr:`unit_conversion_enabled` is
            ``False``
        strict : bool, optional
            Whether to throw an exception if the data in the VTK file being
            read are not formatted in a valid way
        fallback_units : dict, optional
            This setting allows VTK files where some data are missing units
            in the identifier to still be read with unit conversions enabled.
            Must be a dictionary where keys and values are both strings.  If
            :py:attr:`unit_conversion_enabled` is ``True`` and a VTK data
            identifier does not include the data units, then if there is a key
            in ``fallback_units`` matching the identifier, the corresponding
            value in ``fallback_units`` will be set as the units

        Warnings
        --------
        If there are two point data fields in the VTK file with exactly the
        same identifier, only one of the fields (the last one with the
        identifier) will be read.

        Setting ``strict`` to ``True`` modifies the ``stderr`` file descriptor,
        including redirecting ``stderr`` to a temporary file, so it can cause
        problems for other code that relies on streams (for instance, it may
        cause ``unittest`` tests to fail in some cases).
        """
        # SETUP --------------------------------------------------------------
        # Set "path" attribute, verify file exists, and store file hashes
        #   Mypy type annotation added because immediately after calling
        #   `set_read_metadata()`, the "path" attribute cannot be `None`
        #   or else an error would have been thrown
        self.set_read_metadata(path)
        self.path: pathlib.Path

        # Enable/disable unit conversion and verify that "coordinate_units"
        # is provided if and only if unit conversions are enabled
        self._unit_conversion_enabled = bool(unit_conversion_enabled)
        self._check_unit_conversion_compliance_args(coordinate_units)

        # Store units of VTK grid points
        if self.unit_conversion_enabled:
            # Verify that unit doesn't have spaces or other invalid characters
            permissible_chars \
                = string.ascii_letters + string.digits + '()[]{}*/+-^'

            if not pyxx.strings.str_includes_only(str(coordinate_units),
                                                  permissible_chars):
                raise ValueError(
                    f'Unit "{coordinate_units}" contains invalid characters')

            self._coordinate_units = str(coordinate_units)
        else:
            self._coordinate_units = None

        # READ VTK FILE ------------------------------------------------------
        # Set up VTK reader
        reader = vtk.vtkDataSetReader()
        reader.SetFileName(str(self.path))

        # Configure VTK to read all available VTK arrays
        reader.ReadAllFieldsOn()
        reader.ReadAllScalarsOn()
        reader.ReadAllVectorsOn()
        reader.ReadAllTensorsOn()

        # Read data from VTK file
        if strict:
            stderr_stream = io.BytesIO()
            with CaptureStderr(stderr_stream):
                reader.Update()

            if len(stderr_stream.getvalue()) > 0:
                error_message = stderr_stream.getvalue().decode("utf_8")

                raise VTKFormatError(
                    f'VTK file "{self.path}" could not be read successfully. '
                    f'The following error was encountered: {error_message}')
        else:
            reader.Update()

        output = reader.GetOutput()

        # PARSE VTK FILE -----------------------------------------------------
        # Number of points stored in the file and coordinates of each point
        self._num_points = int(output.GetNumberOfPoints())

        x_points: List[float] = []
        y_points: List[float] = []
        z_points: List[float] = []
        for i in range(self.num_points):
            x, y, z = output.GetPoint(i)

            x_points.append(float(x))
            y_points.append(float(y))
            z_points.append(float(z))

        if unit_conversion_enabled:
            df_data = {
                f'x[{self.coordinate_units}]': x_points,
                f'y[{self.coordinate_units}]': y_points,
                f'z[{self.coordinate_units}]': z_points,
            }
        else:
            df_data = {'x': x_points, 'y': y_points, 'z': z_points}

        self.__xyz_coordinate_columns = tuple(df_data.keys())

        self._vtk_data_types \
            = dict(zip(self.__xyz_coordinate_columns,
                       [VTKDataType.scalar] * len(self.__xyz_coordinate_columns)))

        # Read all arrays from VTK file
        point_data = output.GetPointData()

        if self.unit_conversion_enabled:
            data_id_names = [self._parse_column_id(x, 'name')
                             for x in self.__xyz_coordinate_columns]
        else:
            data_id_names = list(self.__xyz_coordinate_columns)

        for i in range(point_data.GetNumberOfArrays()):
            identifier = str(point_data.GetArray(i).GetName())

            if identifier in self.__xyz_coordinate_columns:
                raise VTKIdentifierNameError(
                    f'Invalid VTK data identifier "{identifier}" (matches '
                    'name of one of the point coordinate columns)')

            try:
                self._check_unit_conversion_compliance_id(identifier)
            except VTKIdentifierNameError:
                if (fallback_units is not None) and (identifier in fallback_units):
                    # Add units to data identifier
                    identifier = identifier + f'[{fallback_units[identifier]}]'

                    # Re-check identifier format compliance
                    self._check_unit_conversion_compliance_id(identifier)
                else:
                    raise

            if self.unit_conversion_enabled:
                name = self._parse_column_id(identifier, 'name')

                if name in data_id_names:
                    raise VTKIdentifierNameError(
                        f'Invalid VTK data identifier "{identifier}" (multiple '
                        f'identifiers use the same name "{name}")')

                data_id_names.append(name)

            try:
                array = np.array(
                    vtk.util.numpy_support.vtk_to_numpy(point_data.GetArray(i)),
                    dtype=np.float64)
            except ValueError as exception:
                raise VTKFormatError(
                    f'VTK point data "{identifier}" does not appear to be a '
                    'valid array of numbers') from exception

            # Validate data format and perform pre-processing
            if array.shape[0] != self.num_points:
                raise VTKFormatError(
                    f'VTK point data "{identifier}" has {len(array)} '
                    'elements, which does not match number of VTK grid '
                    f'points ({self.num_points})')

            # Modify data to match expected format for Pandas
            if array.ndim == 1:  # Array contains a scalar for each grid point
                self._vtk_data_types[identifier] = VTKDataType.scalar
                pd_array = list(array)
            elif array.ndim == 2:  # Array contains a vector for each grid point
                if not array.shape[1] == 3:
                    raise VTKFormatError(
                        'VTK vector data should have 3 components (x, y, z), '
                        f'but data specified by "{identifier}" has '
                        f'{array.shape[1]} components')

                self._vtk_data_types[identifier] = VTKDataType.vector
                pd_array = [np.array([v[0], v[1], v[2]]) for v in array]
            else:
                raise VTKFormatError(
                    f'VTK point data specified by "{identifier}" has invalid '
                    f'dimensions {array.shape}. Valid dimensions are:\n'
                    f'-- Scalar data: ({self.num_points},)\n'
                    f'-- Vector data: ({self.num_points}, 3)')

            # Store data
            df_data[identifier] = pd_array

        # CREATE PANDAS DATAFRAME --------------------------------------------
        self._df = pd.DataFrame(df_data)



"""Utilities for capturing text printed to the terminal when running Python
code.
"""

import io
import os
import sys
import tempfile


class CaptureStderr:
    """Captures ``stderr`` text printed to the terminal when running commands

    Redirects output sent to ``stderr``, including output from C-level
    streams.  This approach is adapted from:

    Bendersky (2015). "Redirecting all kinds of stdout in Python." Retrieved from
    https://eli.thegreenplace.net/2015/redirecting-all-kinds-of-stdout-in-python.

    Warnings
    --------
    This class modifies the ``stderr`` file descriptor, including redirecting
    ``stderr`` to a temporary file, so it can cause problems for other code
    that relies on streams (for instance, it may cause ``unittest`` tests to
    fail in some cases).
    """

    def __init__(self, stream: io.BytesIO) -> None:
        # Store stream to which to output captured text
        self._stream = stream

        # Store the original `stderr` file descriptor
        self._current_stderr_fd = sys.stderr.fileno()

        # Create a saved copy of the original `stderr` file descriptor
        self._original_stderr_fd = os.dup(self._current_stderr_fd)

        # Create temporary file to which to redirect `stderr`
        self._temp_file = tempfile.TemporaryFile(mode='w+b')

    def __redirect_stderr(self, file_descriptor: int):
        """Directs ``stderr`` to a given file descriptor"""
        # Flush and close `sys.stderr`
        sys.stderr.close()

        # Direct `stderr` to the given file descriptor
        os.dup2(file_descriptor, self._current_stderr_fd)
        sys.stderr = io.TextIOWrapper(os.fdopen(self._current_stderr_fd, 'wb'))

    def __enter__(self):
        # Redirect `stderr` to a temporary file
        self.__redirect_stderr(self._temp_file.fileno())

    def __exit__(self, *args, **kwargs):
        # Redirect `stderr` back to original file descriptor
        self.__redirect_stderr(self._original_stderr_fd)

        # Write captured `stderr` content to stream
        self._temp_file.flush()
        self._temp_file.seek(0, io.SEEK_SET)
        self._stream.write(self._temp_file.read())

        # Delete temporary file
        self._temp_file.close()

        # Restore original `stderr`
        os.close(self._original_stderr_fd)



"""Classes for representing 2D shapes

The classes in this module are intended to represent arbitrary 2D shapes and
to be used in "drawing" utilities in which complex 2D geometry is assembed
from a variety of simpler shapes.
"""

import math
from typing import Optional, Tuple, TYPE_CHECKING, Union

import numpy as np

from .geometry import Geometry
from .point import Array_Float2

if TYPE_CHECKING:
    from .point2D import CartesianPoint2D  # pragma: no cover


class Shape2D(Geometry):
    """Represents an arbitrary, two-dimensional shape

    This class is intended to represent an arbitrary 2D shape.  Note that
    the intended application for this class is to be used in a drawing utility,
    which is why attributes such as "construction" shapes are defined (these
    are a common feature of computer-aided design tools).
    """

    def __init__(self, is_closed: bool,
                 default_num_coordinates: Optional[int] = None,
                 construction: bool = False,
                 units: Optional[str] = None) -> None:
        """Creates an object representing a 2D geometric shape

        Defines an object which represents a general shape and can return
        shape properties and the Cartesian coordinates of the (discretized)
        shape.

        Parameters
        ----------
        is_closed : bool
            Whether the shape is bounded by a closed perimeter
        default_num_coordinates : int, optional
            The default number of coordinates to use when representing the
            shape (default is ``None``)
        construction : bool, optional
            Whether the shape is a "construction shape" meant for visual
            display but not functional geometry (default is ``False``)
        units : str, optional
            The units in which the geometry is defined, or ``None`` to
            indicate dimensionless geometry or that units are to be ignored
            (default is ``None``)
        """
        super().__init__(units=units)

        self._is_closed = bool(is_closed)
        self.construction = construction
        self.default_num_coordinates = default_num_coordinates

    @property
    def construction(self):
        """Whether the shape is a "construction shape" meant for visual
        display but not functional geometry"""
        return self._construction

    @construction.setter
    def construction(self, construction: bool):
        if not isinstance(construction, bool):
            raise TypeError('Argument "construction" must be of type "bool"')

        self._construction = construction

    @property
    def is_closed(self):
        """Whether the shape is bounded by a closed perimeter"""
        return self._is_closed

    @property
    def default_num_coordinates(self):
        """The number of coordinates to use when representing a discretized
        form of the shape"""
        return self._default_num_coordinates

    @default_num_coordinates.setter
    def default_num_coordinates(self, default_num_coordinates: Optional[int]):
        if default_num_coordinates is None:
            self._default_num_coordinates = None
        else:
            try:
                if not float(default_num_coordinates).is_integer():
                    raise ValueError(
                        'Argument "default_num_coordinates" must be an '
                        'integer, not a floating-point number')
            except (TypeError, ValueError) as exception:
                raise ValueError('Unable to convert "default_num_coordinates" '
                                 'to an integer') from exception

            if int(default_num_coordinates) <= 0:
                raise ValueError(
                    'Argument "default_num_coordinates" must be positive')

            self._default_num_coordinates = int(default_num_coordinates)

    def _convert_rotate_angle(self, angle: float, angle_units: str) -> float:
        if angle_units == 'rad':
            return angle

        if angle_units == 'deg':
            return math.radians(angle)

        raise ValueError('Argument "angle_units" must be either "rad" or "deg"')

    def _convert_xy_coordinates_to_points(self, **kwargs
                                          ) -> Tuple[np.ndarray, ...]:
        x_coordinates, y_coordinates = self.xy_coordinates(**kwargs)
        return tuple(np.transpose(np.array([x_coordinates, y_coordinates])))

    def _get_num_coordinates(self, num_coordinates: Optional[int] = None):
        if num_coordinates is not None:
            return num_coordinates

        if self.default_num_coordinates is not None:
            return self.default_num_coordinates

        raise ValueError('Cannot generate coordinates.  Neither the '
                         '"num_coordinates" argument was be provided nor '
                         'was the "default_num_coordinates" attribute set')

    def points(self) -> Tuple[np.ndarray, ...]:
        """Returns a list containing discretized points around the perimeter
        of the shape

        This method returns a tuple, of which each element is a point along
        the perimeter of the shape.

        See Also
        --------
        xy_coordinates :
            Returns the same coordinates as :py:meth:`xy_coordinates` except
            that points are returned as a list, where each entry is a point on
            the perimeter of the shape (essentially the transpose of
            :py:meth:`xy_coordinates`)
        """
        raise NotImplementedError  # pragma: no cover

    def rotate(self, center: Union[Array_Float2, 'CartesianPoint2D'],
               angle: float, angle_units: str = 'rad') -> None:
        """Rotates the shape in the :math:`xy`-plane

        Rotates the shape the shape a given angle in the :math:`xy`-plane
        about a user-specified point.

        Parameters
        ----------
        center : list or tuple or CartesianPoint2D
            The center of rotation about which to rotate the shape
        angle : float
            The angle by which to rotate the shape about ``center``
        angle_units : str, optional
            The units (radians or degrees) of the ``angle`` argument.  Must be
            either ``'rad'`` or ``'deg'`` (default is ``'rad'``)
        """
        raise NotImplementedError  # pragma: no cover

    def translate(self, x: float = 0, y: float = 0) -> None:
        """Translates the shape in the :math:`xy`-plane

        Translates the shape a user-specified distance in the :math:`x`- and/or
        :math:`y`-directions.

        Parameters
        ----------
        x : float, optional
            The distance to translate the shape in the :math:`x`-direction
            (default is ``0``)
        y : float, optional
            The distance to translate the shape in the :math:`y`-direction
            (default is ``0``)
        """
        raise NotImplementedError  # pragma: no cover

    def xy_coordinates(self) -> Tuple[np.ndarray, np.ndarray]:
        """Generates Cartesian coordinates of the shape

        This method generates a set of discretized points around the perimeter
        of the shape.  Points are returned as a tuple of two NumPy arrays: the
        first NumPy array contains the x-coordinates of the points, and the
        second NumPy array contains the y-coordinates.  This format makes it
        relatively easy to plot the shape using packages like Matplotlib.

        See Also
        --------
        points :
            Returns the same coordinates as :py:meth:`xy_coordinates` except
            that points are returned with the x- and y-coordinates grouped for
            each point (essentially the transpose of :py:meth:`xy_coordinates`)
        """
        raise NotImplementedError  # pragma: no cover



"""Classes for representing a point in space

The classes in this module can be used to store the location of a point in
space.  Although these data could simply be stored in a tuple or list, the
custom classes in this module are advantageous in that they avoid mistakenly
mixing coordinate systems (such as Cartesian and polar) and they provide
added error-checking functionality, such as verifying that coordinates are
floating-point numbers.
"""

from typing import List, Optional, overload, Tuple, Union

import numpy as np

from .geometry import Geometry

# Type alias for a list or tuple containing two floating-point numbers
Array_Float2 = Union[List[float], Tuple[float, float], np.ndarray]

# Type alias for a list or tuple containing three floating-point numbers
Array_Float3 = Union[List[float], Tuple[float, float, float], np.ndarray]


class Point(Geometry):
    """Base class representing an arbitrary point in a space of an arbitrary
    number of dimensions

    This is a generic class that represents an arbitrary point in any
    coordinate system.  In general, this class should be inherited by
    other classes and customizations added (for instance, checking that
    the number of point coordinates match the dimension of the coordinate
    system).

    Notes
    -----
    The equality operator (``==``) is defined for points.  Points are
    considered equal if they are of the same type (:py:class:`Point`,
    :py:class:`CartesianPoint2D`, etc.) and if the points' have
    :py:attr:`coordinates` attributes of the same shape and values.
    """

    def __init__(self, units: Optional[str] = None, **kwargs):
        """Creates an instance of a :py:class:`Point` class and sets the point
        coordinates to an empty tuple

        Parameters
        ----------
        units : str, optional
            The units in which the geometry is defined, or ``None`` to
            indicate dimensionless geometry or that units are to be ignored
            (default is ``None``)
        """
        super().__init__(units=units, **kwargs)

        self._coordinates: Tuple[float, ...] = ()

        # Iterable index
        self.__iter_index = 0

    def __eq__(self, value) -> bool:
        # Verify that `value` is of the same type of point
        if not isinstance(value, self.__class__):
            return False

        # Check that units are the same
        if not self._has_identical_units(value):
            return False

        # If points don't have the same number of coordinates, they aren't equal
        if len(self.coordinates) != len(value.coordinates):
            return False

        # If coordinate values differ, the points aren't equal
        for i, item in enumerate(self.coordinates):
            if item != value.coordinates[i]:
                return False

        return True

    @overload
    def __getitem__(self, index: int) -> float:
        ...  # pragma: no cover

    @overload
    def __getitem__(self, index: slice) -> Tuple[float, ...]:
        ...  # pragma: no cover

    def __getitem__(self, idx):
        return self._coordinates[idx]

    def __iter__(self):
        self.__iter_index = 0
        return self

    def __len__(self):
        return len(self._coordinates)

    def __next__(self):
        if self.__iter_index >= len(self._coordinates):
            self.__iter_index = 0
            raise StopIteration

        self.__iter_index += 1
        return self._coordinates[self.__iter_index - 1]

    def __repr__(self):
        return f'{self.__class__} {self.__str__()}'

    def __str__(self) -> str:
        return str(self._coordinates)

    @property
    def coordinates(self) -> Tuple[float, ...]:
        """The coordinates of the point"""
        return self._coordinates

    @coordinates.setter
    def coordinates(self):
        raise NotImplementedError  # pragma: no cover



"""The classes in this module are intended to store a group of related 2D
shapes.
"""

import itertools
from typing import List, Optional, Union

# Mypy type checking disabled for packages that are not PEP 561-compliant
import plotly.express as px        # type: ignore
import plotly.graph_objects as go  # type: ignore
import pyxx

from mahautils.shapes.geometry.shape_open_closed import ClosedShape2D, Shape2D
from mahautils.shapes.geometry.point import Array_Float2
from mahautils.shapes.geometry.point2D import CartesianPoint2D
from .plotting import _figure_config, _create_blank_plotly_figure


class Layer(pyxx.arrays.TypedListWithID[Shape2D]):
    """An object for storing a set of 2D shapes

    A :py:class:`Layer` is intended to store a set of related shapes, allowing
    more complex, multi-shape geometries to be constructed.  For instance, one
    potential application would be to use a :py:class:`Layer` to store the
    disjoint polygons that constitute the high-pressure ports in an axial
    piston pump.

    The :py:class:`Layer` inherits from Python's ``MutableSequence`` type, so
    most operations that can be performed for a Python ``list`` (append, pop,
    insert, etc.) are valid operations for a :py:class:`Layer`.
    """

    _id = itertools.count(0)

    def __init__(self, *shapes: Shape2D, name: Optional[str] = None,
                 color: str = 'default', print_multiline: bool = True,
                 ) -> None:
        """Creates a new layer to store shapes

        Creates a new :py:class:`Layer` object in which a set of 2D shapes (of
        type :py:class:`Shape2D`) can be stored.

        Parameters
        ----------
        *shapes : Shape2D, optional
            Shapes to add to the layer after the layer is created (default is
            no shapes)
        name : str, optional
            A descriptive name to identify the layer.  If not provided, the
            layer name is generated automatically
        color : str, optional
            The color with which to display the layer in plots.  If not
            provided or set to ``'default'``, the colors from Plotly's
            default color order are used
        print_multiline : bool, optional
            Whether to return a printable string representation of the list in
            multiline format (default is ``True``).  Multiline format places
            each item in the list on its own line

        Notes
        -----
        Any valid color accepted by Plotly can be specified, either as a named
        color (``'blue'``, ``'green'``, etc.) or a hex code.  Valid options are
        described on `this page <https://plotly.com/python/discrete-color/>`__.
        """
        super().__init__(*shapes, list_type=Shape2D,
                         print_multiline=print_multiline, multiline_padding=1)

        self.color = color
        self.name = f'layer{self.id}' if name is None else name

    @property
    def color(self) -> str:
        """The color with which to display the layer in plots

        Notes
        -----
        Any valid color accepted by Plotly can be specified, either as a named
        color (``'blue'``, ``'green'``, etc.) or a hex code.  Valid options are
        described on `this page <https://plotly.com/python/discrete-color/>`__.
        """
        return self._color

    @color.setter
    def color(self, color: str):
        # Get default Plotly color sequence
        plotly_colors: List[str] = px.colors.qualitative.Plotly

        # Set layer color
        if color == 'default':
            self._color = plotly_colors[self.id % len(plotly_colors)]
        else:
            self._color = str(color)

    @property
    def name(self) -> str:
        """A descriptive name to identify the layer"""
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = str(name)

    @property
    def num_shapes(self) -> int:
        """The number of shapes in the layer"""
        return len(self)

    def plot(self, units: Optional[str] = None,
             figure: Optional[go.Figure] = None,
             show: bool = True, return_fig: bool = False,
             ) -> Union[go.Figure, None]:
        """Plots the shapes in the layer

        Creates a Plotly figure illustrating the shapes in the layer, or
        optionally appends traces for each shape to an existing figure.  The
        figure can be opened in a browser (default behavior) and/or returned
        (to allow subsequent user-specific customizations).

        Parameters
        ----------
        units : str, optional
            If units are provided, it will be verified that all plotted shapes
            have these units in their :py:attr:`Shape2D.units` attribute
            (default is ``None``, which performs no unit checks).  Additionally,
            if the ``figure`` argument was not provided, the specified units
            will be included in the axis titles
        figure : go.Figure, optional
            A Plotly figure.  If provided, rather than creating a new figure
            from scratch, the layer's shapes will be added as new traces in
            the provided figure (default is ``None``, which creates a new
            figure from scratch)
        show : bool, optional
            Whether to open the figure in a browser (default is ``True``)
        return_fig : bool, optional
            Whether to return the figure (default is ``False``)

        Returns
        -------
        go.Figure
            A Plotly figure depicting the layer.  Returned if and only if
            ``return_fig`` is ``True``
        """
        if not isinstance(figure, go.Figure):
            figure = _create_blank_plotly_figure(units)

        for shape in self:
            if (units is not None) and (units != shape.units):
                raise ValueError(
                    f'Expected all shapes to have units "{units}" but found '
                    f'a shape with units "{shape.units}"')

            if isinstance(shape, ClosedShape2D):
                x, y = shape.xy_coordinates(repeat_end=True)

                if not shape.construction:
                    figure.add_trace(go.Scatter(
                        x=x, y=y, fill='toself',
                        fillcolor=self.color, line=None, opacity=0.2,
                        hoverinfo='skip',
                    ))
            else:
                x, y = shape.xy_coordinates()

            figure.add_trace(go.Scatter(
                x=x, y=y, fill=None, opacity=1, fillcolor=None,
                mode='lines' if shape.construction else 'lines+markers',
                line={
                    'color': self.color,
                    'dash': 'dash' if shape.construction else 'solid',
                },
                marker={'size': 4},
                hovertemplate='(%{x}, %{y})<extra></extra>',
            ))

        if show:
            figure.show(config=_figure_config)

        if return_fig:
            return figure

        return None

    def rotate(self, center: Union[Array_Float2, CartesianPoint2D],
               angle: float, angle_units: str = 'rad') -> None:
        """Rotates all shapes in the layer a given angle in the :math:`xy`-plane
        about a user-specified point

        Parameters
        ----------
        center : list or tuple or CartesianPoint2D
            The center of rotation about which to rotate the shapes
        angle : float
            The angle by which to rotate the shapes about ``center``
        angle_units : str, optional
            The units (radians or degrees) of the ``angle`` argument.  Must be
            either ``'rad'`` or ``'deg'`` (default is ``'rad'``)
        """
        for shape in self:
            shape.rotate(center=center, angle=angle, angle_units=angle_units)

    def translate(self, x: float = 0, y: float = 0) -> None:
        """Translates all shapes in the layer a user-specified distance in
        the :math:`x`- and/or :math:`y`-directions

        Parameters
        ----------
        x : float, optional
            The distance to translate the shapes in the :math:`x`-direction
            (default is ``0``)
        y : float, optional
            The distance to translate the shapes in the :math:`y`-direction
            (default is ``0``)
        """
        for shape in self:
            shape.translate(x=x, y=y)



"""Classes for representing a circle

The classes in this module are intended to represent a circle in the 2D
Cartesian coordinate system.
"""

import math
from typing import Optional, Tuple, Union

import numpy as np

from .point import Array_Float2
from .point2D import CartesianPoint2D
from .shape_open_closed import ClosedShape2D


class Circle(ClosedShape2D):
    """An object representing a circle in the 2D Cartesian plane

    This class is intended to represent a circle in 2D Cartesian coordinates.
    Users are required to define the circle's center and radius (or diameter)
    when creating :py:class:`Circle` objects, and the objects can then perform
    actions such as returning discretized points on the circumference of the
    circle and determining whether a point is inside the circle.

    Examples
    --------
    Create a circle centered at :math:`(0,0)` with radius :math:`2`:

    >>> circle = mahautils.shapes.Circle((0,0), radius=2)

    Generate points around the circle's circumference (output is rounded here
    for more easily understandable display):

    >>> import numpy as np
    >>> np.round(circle.points(num_coordinates=4, repeat_end=False))
    array([[ 2.,  0.],
           [ 0.,  2.],
           [-2.,  0.],
           [-0., -2.]])

    Check whether a point is inside the circle:

    >>> circle.is_inside((0, 0))
    True
    >>> circle.is_inside((2.01, 0))
    False
    """

    def __init__(self, center: Union[Array_Float2, CartesianPoint2D],
                 radius: Optional[float] = None, diameter: Optional[float] = None,
                 default_num_coordinates: Optional[int] = None,
                 construction: bool = False,
                 polygon_file_enclosed_conv: int = 1,
                 units: Optional[str] = None) -> None:
        """Creates an object representing a circle

        Defines a circle in the 2D Cartesian plane, locating it based on the
        center of the circle and the radius (or diameter).

        Parameters
        ----------
        center : list or tuple or CartesianPoint2D
            The center of the circle
        radius : float, optional
            The radius of the circle (default is ``None``)
        diameter : float, optional
            The diameter of the circle (default is ``None``)
        default_num_coordinates : int, optional
            The default number of coordinates to use when representing the
            shape (default is ``None``)
        construction : bool, optional
            Whether the shape is a "construction shape" meant for visual
            display but not functional geometry (default is ``False``)
        polygon_file_enclosed_conv : int, optional
            Convention for considering enclosed area when generating a polygon
            file from :py:class:`ClosedShape2D` objects (default is ``1``)
        units : str, optional
            The units in which the geometry is defined, or ``None`` to
            indicate dimensionless geometry or that units are to be ignored
            (default is ``None``)

        Notes
        -----
        It is required that either ``radius`` or ``diameter`` is provided, but
        not both.  If neither or both are provided, a :py:class:`TypeError`
        will be thrown.
        """
        super().__init__(
            default_num_coordinates=default_num_coordinates,
            construction=construction,
            polygon_file_enclosed_conv=polygon_file_enclosed_conv,
            units=units,
        )

        # Store circle center.  Mypy is disabled for this line because it
        # incorrectly identified the types of the argument and the attribute
        # as incompatible (even though they should be since the argument and
        # the setter function accept the same types)
        self.center = center  # type: ignore

        # Store circle radius or diameter
        if sum([radius is None, diameter is None]) != 1:
            raise TypeError(
                'Exactly one of the following arguments must be provided: '
                '["radius", "diameter"]')

        if radius is not None:
            self.radius = radius
        else:
            # Mypy is disabled on this line -- from previous check, "diameter"
            # cannot be "None" since it was already confirmed that either
            # "radius" or "diameter" is not "None", and this "else" statement
            # is only entered if "radius" is "None"
            self.diameter = diameter  # type: ignore

    def __eq__(self, value: object) -> bool:
        # Check that operand is of type "Circle"
        if not isinstance(value, Circle):
            return False

        # Check that units are the same
        if not self._has_identical_units(value):
            return False

        # If circles have the same center and radius, they are considered equal
        if (self.center == value.center) and (self.radius == value.radius):
            return True

        return False

    def __repr__(self):
        return f'{__class__} center={self.center}, radius={self.radius}'

    def __str__(self):
        return self.__repr__()

    @property
    def center(self) -> CartesianPoint2D:
        """The center of the circle"""
        return self._center

    @center.setter
    def center(self, center: Union[Array_Float2, CartesianPoint2D]):
        self._center = CartesianPoint2D(center)

    @property
    def diameter(self) -> float:
        """The diameter of the circle"""
        return 2 * self.radius

    @diameter.setter
    def diameter(self, diameter: float):
        self.radius = diameter / 2

    @property
    def radius(self) -> float:
        """The radius of the circle"""
        return self._radius

    @radius.setter
    def radius(self, radius: float):
        if float(radius) < 0:
            raise ValueError('Circle radius cannot be negative')

        self._radius = float(radius)

    def is_inside(self, point: Union[Array_Float2, CartesianPoint2D],
                  perimeter_is_inside: bool = True) -> bool:
        distance = self.center.distance_to(point)

        if perimeter_is_inside:
            return distance <= self.radius
        return distance < self.radius

    def points(self, repeat_end: bool = False,
               num_coordinates: Optional[int] = None,
               ) -> Tuple[np.ndarray, ...]:
        """Returns a list containing discretized points around the
        circumference of the circle

        This method returns a tuple, of which each element is a point along
        the circumference of the circle.  Points are ordered in a
        counterclockwise direction around the circle.

        Parameters
        ----------
        repeat_end : bool, optional
            Whether the first and last coordinate returned should be the same
            point (default is ``False``).  This is useful, for instance, when
            plotting the circle with Matplotlib: if ``repeat_end`` is set to
            ``False``, there may be a slight gap visible between the end points
            of the circle
        num_coordinates : int, optional
            The number of points to use when discretizing the circle's shape
            (default is ``None``).  If this argument is ``None`` or omitted,
            the number of points is set to :py:attr:`default_num_coordinates`,
            if provided (otherwise an error is thrown)

        See Also
        --------
        points :
            Returns the same coordinates as :py:meth:`points` except
            that points are returned with all the x-coordinates aggregated and
            all the y-coordinates aggregated (essentially the transpose of
            :py:meth:`points`)
        """
        return self._convert_xy_coordinates_to_points(
            num_coordinates=num_coordinates,
            repeat_end=repeat_end,
        )

    def rotate(self, center: Union[Array_Float2, CartesianPoint2D],
               angle: float, angle_units: str = 'rad') -> None:
        self.center.rotate(center=center, angle=angle, angle_units=angle_units)

    def translate(self, x: float = 0, y: float = 0) -> None:
        self.center.translate(x=x, y=y)

    def xy_coordinates(self, repeat_end: bool = False,
                       num_coordinates: Optional[int] = None,
                       ) -> Tuple[np.ndarray, np.ndarray]:
        """Generates Cartesian coordinates of the circle

        This method generates a set of discretized points around the
        circumference of the circle, in counterclockwise order.  Points are
        returned as a tuple of two NumPy arrays: the first NumPy array
        contains the x-coordinates of the points, and the second NumPy array
        contains the y-coordinates.  This format makes it relatively easy to
        plot the shape using packages like Matplotlib.

        Parameters
        ----------
        repeat_end : bool, optional
            Whether the first and last coordinate returned should be the same
            point (default is ``False``).  This is useful, for instance, when
            plotting the circle with Matplotlib: if ``repeat_end`` is set to
            ``False``, there may be a slight gap visible between the end points
            of the circle
        num_coordinates : int, optional
            The number of points to use when discretizing the circle's shape
            (default is ``None``).  If this argument is ``None`` or omitted,
            the number of points is set to :py:attr:`default_num_coordinates`,
            if provided (otherwise an error is thrown)

        See Also
        --------
        points :
            Returns the same coordinates as :py:meth:`xy_coordinates` except
            that points are returned with the x- and y-coordinates grouped for
            each point (essentially the transpose of :py:meth:`xy_coordinates`)

        Examples
        --------
        Create a :py:class:`Circle` object and plot it with Matplotlib:

        >>> import matplotlib.pyplot as plt
        >>> circle = mahautils.shapes.Circle((0,0), radius=2)
        >>> fig = plt.plot(*circle.xy_coordinates(num_coordinates=1000))
        """
        # Determine the number of coordinates to output
        n = self._get_num_coordinates(num_coordinates)

        # Set the angles for which to output circle coordinates
        if repeat_end:
            end = 2 * math.pi
        else:
            end = 2 * math.pi * ((n - 1) / n)
        theta = np.linspace(0, end, n)

        # Calculate and return circle coordinates
        x_coordinates = self.radius * np.cos(theta) + self.center.x
        y_coordinates = self.radius * np.sin(theta) + self.center.y

        return (x_coordinates, y_coordinates)
